/*
 * unity-webapps-action-manager.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib/gstdio.h>
#include <gio/gio.h>

#include <libdbusmenu-glib/menuitem.h>
#include <libdbusmenu-glib/server.h>

#include "unity-webapps-action-manager.h"

#include "../unity-webapps-debug.h"

struct _UnityWebappsActionManagerPrivate {
  DbusmenuServer *menu_server;
  DbusmenuMenuitem *menu_item;
  
  GHashTable *paths_by_menu_item;
};


G_DEFINE_TYPE(UnityWebappsActionManager, unity_webapps_action_manager, G_TYPE_OBJECT)

enum {
  ACTION_INVOKED,
  LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static UnityWebappsActionManager *default_manager = NULL;

static DbusmenuMenuitem * add_menuitem_by_action_path (UnityWebappsActionManager *manager, const gchar *path);

#define UNITY_WEBAPPS_ACTION_MANAGER_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), UNITY_WEBAPPS_TYPE_ACTION_MANAGER, UnityWebappsActionManagerPrivate))

static gchar **
split_action_path (const gchar *action_path)
{
  gchar **components;
  
  if (action_path[0] != '/')
    {
      return NULL;
    }
  
  components = g_strsplit (action_path+1, "/", -1);
  
  if (g_strv_length (components) > 3)
    {
      g_strfreev (components);
      return NULL;
    }
  
  return components;
}

static void
on_menuitem_activated (DbusmenuMenuitem *item,
		       guint timestamp,
		       gpointer user_data)
{
  UnityWebappsActionManager *action_manager;
  const gchar *path;
  
  action_manager = (UnityWebappsActionManager *)user_data;
  
  path = g_hash_table_lookup (action_manager->priv->paths_by_menu_item, item);
  
  if (path == NULL)
    {
      g_warning("Warning: Got a menu item activation (%p), which we do not have a path for.", item);
    }
  
  g_signal_emit (action_manager, signals[ACTION_INVOKED], 0, path);
}

static DbusmenuMenuitem *
menuitem_find_child_by_label (DbusmenuMenuitem *item,
			      const gchar *label)
{
  const GList *children;
  const GList *walk;
  
  children = dbusmenu_menuitem_get_children (item);
  
  for (walk = children;
       walk != NULL;
       walk = walk->next)
    {
      DbusmenuMenuitem *child;
      
      const gchar *child_label;
      
      child = (DbusmenuMenuitem *)walk->data;
      child_label = dbusmenu_menuitem_property_get (child, "label");
      
      if (g_strcmp0 (child_label, label) == 0)
	{
	  return child;
	}
    }
  
  return NULL;
}

static DbusmenuMenuitem *
add_menuitem_by_name (UnityWebappsActionManager *manager,
		      DbusmenuMenuitem *parent,
		      const gchar *name)
{
  DbusmenuMenuitem *action_item;

  if (menuitem_find_child_by_label (parent, name) != NULL)
    {
      return menuitem_find_child_by_label (parent, name);
    }
  
  action_item = dbusmenu_menuitem_new ();

  dbusmenu_menuitem_property_set (action_item, DBUSMENU_MENUITEM_PROP_LABEL, name);
  dbusmenu_menuitem_property_set_bool (action_item, DBUSMENU_MENUITEM_PROP_VISIBLE, TRUE);
  
  g_signal_connect (action_item, "item-activated", G_CALLBACK(on_menuitem_activated), manager);
  
  dbusmenu_menuitem_child_append (parent, action_item);
  
  return action_item;
}

static DbusmenuMenuitem *
add_menuitem_by_action_path (UnityWebappsActionManager *action_manager,
		      const gchar *path)
{
  DbusmenuMenuitem *parent_menuitem;
  gchar **components;
  gint i, length;
  DbusmenuMenuitem *ret;
  
  components = split_action_path (path);
  
  if (components == NULL)
    {
      return NULL;
    }

  parent_menuitem = action_manager->priv->menu_item;
  
  length = g_strv_length (components);
  
  ret = NULL;
  
  for (i = 0; i < length -1; i++)
    {
      DbusmenuMenuitem *new_parent;
      gchar *component;
      
      component = components[i];
      
      new_parent = menuitem_find_child_by_label (parent_menuitem, component);
      
      if (new_parent == NULL)
	{
	  new_parent = add_menuitem_by_name (action_manager, parent_menuitem,
					  component);
	  
	  if (new_parent == NULL)
	    {
	      goto out;
	    }
	  
	}
      
      parent_menuitem = new_parent;
    }
  
  ret = add_menuitem_by_name (action_manager, parent_menuitem, components[i]);
  
 out:
  g_strfreev (components);
  
  return ret;
}

static void
unity_webapps_action_manager_finalize (GObject *object)
{
  UnityWebappsActionManager *manager;
  
  manager = UNITY_WEBAPPS_ACTION_MANAGER (object);
  
  g_hash_table_destroy (manager->priv->paths_by_menu_item);
 
  g_object_unref (G_OBJECT (manager->priv->menu_item));
  g_object_unref (G_OBJECT (manager->priv->menu_server));
}

static void
unity_webapps_action_manager_class_init (UnityWebappsActionManagerClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  
  object_class->finalize = unity_webapps_action_manager_finalize;
  
  g_type_class_add_private (object_class, sizeof(UnityWebappsActionManagerPrivate));
  
  signals[ACTION_INVOKED] = g_signal_new ("action-invoked",
					  UNITY_WEBAPPS_TYPE_ACTION_MANAGER,
					  G_SIGNAL_RUN_LAST,
					  0,
					  NULL,
					  NULL,
					  NULL,
					  G_TYPE_NONE,
					  1,
					  G_TYPE_STRING);
}

static void
unity_webapps_action_manager_init (UnityWebappsActionManager *manager)
{
  manager->priv = UNITY_WEBAPPS_ACTION_MANAGER_GET_PRIVATE (manager);
  
  manager->priv->menu_server = dbusmenu_server_new ("/com/canonical/Unity/Webapps/Context/ApplicationActions");
  manager->priv->menu_item = dbusmenu_menuitem_new ();
  
  dbusmenu_server_set_root (manager->priv->menu_server, manager->priv->menu_item);
  
  manager->priv->paths_by_menu_item = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, g_free);
}

UnityWebappsActionManager *
unity_webapps_action_manager_new ()
{
  return g_object_new (UNITY_WEBAPPS_TYPE_ACTION_MANAGER, NULL);
}

UnityWebappsActionManager *
unity_webapps_action_manager_get_default ()
{
  if (default_manager == NULL)
    {
      default_manager = g_object_new (UNITY_WEBAPPS_TYPE_ACTION_MANAGER, NULL);
    }
  return default_manager;
}

void
unity_webapps_action_manager_add_action (UnityWebappsActionManager *manager,
					 const gchar *path)
{
  DbusmenuMenuitem *added;
  
  added = add_menuitem_by_action_path (manager, path);
  
  if (added == NULL)
    {
      g_warning ("Failed to add action: %s", path);
      
      return;
    }
  
  g_hash_table_insert (manager->priv->paths_by_menu_item, (gpointer)added, g_strdup (path));
}

