/*
 * unity-webapps-context-daemon.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdlib.h>

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gtk/gtk.h>

#include "unity-webapps-context-daemon.h"
#include "../unity-webapps-gen-context.h"

#include "../unity-webapps-dbus-defs.h"

#include "unity-webapps-dirs.h"
#include "unity-webapps-resource-db.h"
#include "unity-webapps-icon-theme.h"

#include "../unity-webapps-debug.h"

#include "unity-webapps-interest-manager.h"
#include "unity-webapps-action-manager.h"

#include "unity-webapps-wire-version.h"


/* 
 * If a local icon file has been downloaded more recently than this time (in seconds) we will simply use our existing
 * download. In the future maybe this should be more intelligent and use HTTP IfModifiedSince 
 *
 */
#define UNITY_WEBAPPS_CONTEXT_ICON_STALE_TIME 60 /*seconds*/ * 30 /*minutes*/

static GMainLoop *mainloop = NULL;

static GDBusConnection *session_bus = NULL;

static UnityWebappsGenContext *context_service_interface;

static gchar *
get_icon_name (const gchar *name,
	       const gchar *domain);

static void
export_object (GDBusConnection *connection,
	       UnityWebappsContextDaemon *daemon);

static void
emit_context_ready (GDBusConnection *connection);
static void
emit_no_interest (GDBusConnection *connection);
static void
emit_interests_changed (GDBusConnection *connection);

static GQueue *most_recent_interest_queue = NULL;

void unity_webapps_context_daemon_emit_raise (GDBusConnection *connection, gint interest_id);
void unity_webapps_context_daemon_emit_view_is_active_changed (GDBusConnection *connection, gint interest_id, gboolean is_active);
void unity_webapps_context_daemon_emit_view_location_changed (GDBusConnection *connection, gint interest_id, const gchar *location);
void unity_webapps_context_daemon_emit_view_window_changed (GDBusConnection *connection, gint interest_id, guint64 window);
void unity_webapps_context_daemon_emit_preview_requested (GDBusConnection *connection, gint interest_id);
void unity_webapps_context_daemon_emit_action_invoked (GDBusConnection *connection, const gchar *name);

typedef struct _UnityWebappsContextInfo {
  gchar *name;
  gchar *domain;

  gchar *icon_url;
  
  gboolean icon_is_themed;
  gchar *icon_name;
} UnityWebappsContextInfo;

static UnityWebappsContextInfo info;

#define VALID_NAME_CHARACTERS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"

static void
emit_no_interest (GDBusConnection *connection)
{
  GError *error;
  
  error = NULL;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting NoInterest signal");

  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "NoInterest",
				 NULL, /* params */
				 &error /* Error */);
  
  if (error != NULL)
    {
      /*
       * TODO: Should this be fatal? If this fails we can not guarantee that the daemon will ever close...
	  */
      g_critical ("Error emitting NoInterest signal (from context daemon): %s", error->message);
      g_error_free (error);
      return;
    }
}

static void
emit_context_ready (GDBusConnection *connection)
{
  GError *error;
  
  error = NULL;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting ContextReady signal");
  
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "ContextReady",
				 g_variant_new ("(s)", UNITY_WEBAPPS_WIRE_PROTOCOL_VERSION, NULL),
				 &error);
  /* 
   * TODO:
   * Maybe this should be a fatal error instead of a critical.
   * In this case the service isn't managing the daemon, and so we can't be assured
   * the daemon will ever quit.
   */
  if (error != NULL)
    {
      g_critical("Error emiting ContextReady signal: %s", error->message);
      g_error_free (error);
      
      return;
    }
}

static void
emit_interests_changed (GDBusConnection *connection)
{
  GError *error;
  
  error = NULL;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting InterestsChanged signal");
  
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "InterestsChanged",
				 NULL /* Params */,
				 &error);
  /* 
   * TODO:
   * Maybe this should be a fatal error instead of a critical.
   * In this case the service isn't managing the daemon, and so we can't be assured
   * the daemon will ever quit.
   */
  if (error != NULL)
    {
      g_critical("Error emiting InterestsChanged signal: %s", error->message);
      g_error_free (error);
      
      return;
    }
}

/* 
 * Get a canoncical name string (uniquely identifying the application) which is safe to be used for DBus paths
 * file names, etc (containing only characters from VALID_NAME_CHARACTERS).
 */
static gchar *
get_canonical_full_name (const gchar *name, const gchar *domain)
{
  gchar *canonical_name, **components;

  canonical_name = g_strdup_printf ("%s%s", name, domain);
  g_strcanon (canonical_name, VALID_NAME_CHARACTERS, '%');


  components = g_strsplit (canonical_name, "%", 0);

  g_free (canonical_name);

  canonical_name = g_strjoinv (NULL, components);

  g_strfreev(components);

  return canonical_name;
}

/* Note we need to be more intelligent about this currently we wont write a new desktop file if we are given an icon when we didnt have one before */
static void
write_desktop_file (UnityWebappsContextDaemon *daemon)
{
  GError *error;
  gchar *path, *contents, *desktop_file_name, *t, *icon_file_name, *applications_dir;
  
  
  t = get_canonical_full_name (info.name, info.domain);
  desktop_file_name = g_strdup_printf("%s.desktop", t);
  g_free (t);
  
  applications_dir = unity_webapps_dirs_get_application_dir ();
  path = g_build_filename (applications_dir, desktop_file_name, NULL);

  
  if (info.icon_is_themed)
    {
      icon_file_name = g_strdup (info.icon_name);
    }
  else if (daemon->icon_file_name)
    {
      icon_file_name = get_icon_name (info.name, info.domain);
    }
  else
    {
      icon_file_name = g_strdup ("application-default-icon");
    }
  /* 
   * Eventually we should write mimetypes and an exec entry. The exec entry will point to a small application
   * which simply calls the ActivateApplication method of the webapps serice.
   */
  contents = g_strdup_printf("[Desktop Entry]\nName=%s\nType=Application\nIcon=%s\nExec=unity-webapps-runner -n '%s' -d '%s'", info.name, icon_file_name, info.name, info.domain);

  
  error = NULL;
  
  g_file_set_contents (path, contents, -1, &error);
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Wrote desktop file: %s", path);
  
  if (error)
    {
      g_critical ("Error writing desktop file (%s) contents: %s", path, error->message);
      g_error_free (error);
      
    }
  
  g_free (desktop_file_name);
  g_free (path);
  g_free (icon_file_name);
  g_free (contents);
  g_free (applications_dir);
}

static gchar *
get_desktop_file_name (UnityWebappsContextDaemon *daemon)
{
  gchar *path, *desktop_file_name, *t, *applications_dir;

  t = get_canonical_full_name (info.name, info.domain);
  desktop_file_name = g_strdup_printf("%s.desktop", t);
  g_free (t);
  
  applications_dir = unity_webapps_dirs_get_application_dir ();
  path = g_build_filename (applications_dir, desktop_file_name, NULL);

  //  if (g_file_test(path, G_FILE_TEST_EXISTS) == FALSE)
    {
      write_desktop_file(daemon);
    }
  g_free (path);
  g_free (applications_dir);
  
  return desktop_file_name;

}


static void
bus_filter_service_ready (GDBusConnection *connection,
			  const gchar *sender,
			  const gchar *object,
			  const gchar *interface,
			  const gchar *signal,
			  GVariant *params,
			  gpointer user_data)
{

  if (g_strcmp0(signal, "ServiceReady") != 0)
    {
      g_warning ("Unexpected signal on context daemon instance: %s", signal);
      return;
    }

  /* 
   * If we have received the ServiceReady signal, it means that the service has crashed.
   * So, we should emit the ContextReady signal again, such that the new service daemon
   * can track the context.
   */
  UNITY_WEBAPPS_NOTE (CONTEXT, "Got ServiceReady signal");
  emit_context_ready (session_bus);
}

static gint
add_interest (GDBusConnection *connection,
	      UnityWebappsContextDaemon *daemon,
	      const gchar *name)
{
  UnityWebappsInterestManager *manager;
  
  manager = unity_webapps_interest_manager_get_default ();
  
  return unity_webapps_interest_manager_add_interest (manager, name);
}

/*
 * A client has explicitly removed it's interest. If we have interests we need to emit the signal.
 */
 
static void
lost_interest (GDBusConnection *connection,
	       UnityWebappsContextDaemon *daemon,
	       gint interest_id)
{
  UnityWebappsInterestManager *manager;
  
  manager = unity_webapps_interest_manager_get_default ();
  
  unity_webapps_interest_manager_remove_interest (manager, interest_id);
}

static void
bus_get_cb (GObject *object, GAsyncResult *res, gpointer user_data)
{
  GError *error = NULL;
  GDBusConnection *connection = g_bus_get_finish(res, &error);
  UnityWebappsContextDaemon *daemon;
 
  
  if (error != NULL) 
    {
      g_error("Unable to get session bus: %s", error->message);
      g_error_free(error);
      return;
    }
  
  session_bus = connection;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Connected to the Session Bus");
  
  // TODO: Investigate free on shutdown, just for correctness.
  daemon = unity_webapps_context_daemon_new ();
  
  export_object (session_bus, daemon);

  emit_context_ready (session_bus);
  
  g_dbus_connection_signal_subscribe (session_bus,
				      NULL,
				      UNITY_WEBAPPS_SERVICE_IFACE,
				      "ServiceReady",
				      NULL,
				      NULL,
				      G_DBUS_SIGNAL_FLAGS_NONE,
				      bus_filter_service_ready,
				      daemon /* user data */,
				      NULL /* destroy notify */);
				 
}

static gboolean
parse_args (gint argc, gchar **argv)
{
  if (argc != 3 && argc != 4)
    {
    return FALSE;
    }
  /*
   * TODO: Should we require an icon url???
   */
  else if (argc == 3)
    {
      info.name = argv[1];
      info.domain = argv[2];
      info.icon_url = NULL;
    }
  else if (argc == 4)
    {
      info.name = argv[1];
      info.domain = argv[2];
      
      if (g_strcmp0 (argv[3], "none") == 0)
	{
	  info.icon_url = NULL;
	}
      else
	{
	  info.icon_url = argv[3];
	  
	  if (g_str_has_prefix(info.icon_url, "icon://"))
	    {
	      info.icon_is_themed = TRUE;
	      info.icon_name = info.icon_url + 7 * sizeof (gchar);
	    }
	  
	}
    }
      
  return TRUE;
}

gint 
main (gint argc, gchar **argv)
{
  gtk_init (&argc, &argv);
  
  if (!parse_args (argc, argv))
    return 1;
  
#ifdef UNITY_WEBAPPS_ENABLE_DEBUG
  unity_webapps_debug_initialize_flags ();
#endif
  
  
  unity_webapps_resource_db_open ();
  
  g_bus_get(G_BUS_TYPE_SESSION,
	    NULL, /* Cancellable */
	    bus_get_cb,
	    NULL /* User data */);
  mainloop = g_main_loop_new(NULL, FALSE);
  g_main_loop_run(mainloop);
  

  return 0;
}

/*
 * TODO: Should this be the same as the desktop name?
 */
static gchar *
get_icon_name (const gchar *name,
	       const gchar *domain)
{
  gchar *canonical_name, **components, *icon_name;
  
  canonical_name = g_strdup (name);
  g_strcanon (canonical_name, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890", '%');
  
  components = g_strsplit (canonical_name, "%", 0);
  
  g_free (canonical_name);
  
  canonical_name = g_strjoinv (NULL, components);
  
  icon_name = g_strdup_printf("%s-%s", canonical_name, domain);
  
  g_strfreev (components);
  g_free (canonical_name);
  
  return icon_name;
}

static gchar *
save_icon_file (const gchar *name,
		const gchar *domain,
		const gchar *icon_url)
{
  gchar *icon_name, *icon_full_name;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Saving icon file: %s", icon_url);
  
  icon_name = get_icon_name (name, domain);
  
  icon_full_name = unity_webapps_icon_theme_update_application_icon (icon_name,
								     icon_url);
  
  g_free (icon_name);
  
  return icon_full_name;
}

static gboolean
on_handle_add_icon (UnityWebappsGenContext *context,
		    GDBusMethodInvocation *invocation,
		    const gchar *icon_url,
		    gint size,
		    gpointer user_data)
{
  gchar *icon_name;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling AddIcon call %s with size %d", icon_url, size);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  icon_name = get_icon_name (info.name, info.domain);
  
  unity_webapps_icon_theme_add_icon (icon_name, icon_url, size);
  
  g_free (icon_name);
  
  return TRUE;
}

static gboolean
on_handle_get_icon_name (UnityWebappsGenContext *context,
			 GDBusMethodInvocation *invocation,
			 gpointer user_data)
{
  UnityWebappsContextDaemon *daemon;
  gchar *icon_file_name;
  
  daemon = (UnityWebappsContextDaemon *)user_data;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling GetIconName call");
  
  if (info.icon_is_themed)
    {
      icon_file_name = g_strdup (info.icon_name);
    }
  else if (daemon->icon_file_name)
    {
      icon_file_name = get_icon_name (info.name, info.domain);
    }
  else
    {
      icon_file_name = g_strdup ("application-default-icon");
    }
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Icon name is: %s", icon_file_name);
  
  g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", icon_file_name, NULL));
  
  g_free (icon_file_name);

  return TRUE;
}

static gboolean
on_handle_add_interest (UnityWebappsGenContext *context,
			GDBusMethodInvocation *invocation,
			gpointer user_data)
{
  UnityWebappsContextDaemon *context_daemon;
  const gchar *fan;
  gint interest_id;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling AddInterest call");
  
  context_daemon = (UnityWebappsContextDaemon *)user_data;
  
  fan = g_dbus_method_invocation_get_sender (invocation);
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Interest is from %s", fan);

  interest_id = add_interest (session_bus, context_daemon, fan);
  
  g_dbus_method_invocation_return_value (invocation, g_variant_new ("(i)", interest_id, NULL));
  
  return TRUE;
}

static gboolean
on_handle_lost_interest (UnityWebappsGenContext *context,
			 GDBusMethodInvocation *invocation,
			 gint interest_id,
			 gpointer user_data)
{
  UnityWebappsContextDaemon *context_daemon;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling LostInterest call");

  context_daemon = (UnityWebappsContextDaemon *)user_data;
  
  lost_interest (session_bus, context_daemon, interest_id);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_list_interests (UnityWebappsGenContext *context,
			  GDBusMethodInvocation *invocation,
			  gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  GVariant *interests_list;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling ListInterests call");
  
  manager = unity_webapps_interest_manager_get_default ();
  
  interests_list = unity_webapps_interest_manager_list_interests (manager);
  
  if (interests_list == NULL)
    {
      GVariant *array;
      
      array = g_variant_new_array (G_VARIANT_TYPE_INT32, NULL, 0);
      g_dbus_method_invocation_return_value (invocation, g_variant_new_tuple (&array, 1));
    }
  else
    {
      g_dbus_method_invocation_return_value (invocation, g_variant_new_tuple (&interests_list, 1));
    }
  
  return TRUE;
}

static gboolean
on_handle_get_interest_count (UnityWebappsGenContext *context,
			      GDBusMethodInvocation *invocation,
			      gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  gint num_interests;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling GetInterestCount call");
  
  manager = unity_webapps_interest_manager_get_default ();
  
  num_interests = (gint) unity_webapps_interest_manager_get_num_interests (manager);
  
  g_dbus_method_invocation_return_value (invocation,
					 g_variant_new ("(i)", num_interests, NULL));
  UNITY_WEBAPPS_NOTE (CONTEXT, "Number of interests is %d", num_interests);
  
  return TRUE;

}

static gboolean
on_handle_get_interest_owner (UnityWebappsGenContext *context,
			      GDBusMethodInvocation *invocation,
			      gint interest_id,
			      gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  const gchar *owner;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling GetInterestOwner call");
  
  manager = unity_webapps_interest_manager_get_default ();
  
  owner = unity_webapps_interest_manager_get_interest_owner (manager, interest_id);
  
  if (owner == NULL)
    {
      g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", "", NULL));
      UNITY_WEBAPPS_NOTE (CONTEXT, "Interest not found (%d)", interest_id);
    }
  else
    {
      g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", owner, NULL));
      UNITY_WEBAPPS_NOTE (CONTEXT, "Interest Owner is: %s", owner);
    }
  
  return TRUE;
}

static gboolean
on_handle_get_interest_is_active (UnityWebappsGenContext *context,
				  GDBusMethodInvocation *invocation,
				  gint interest_id,
				  gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  gboolean is_active;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling GetInterestIsActive call");
  
  manager = unity_webapps_interest_manager_get_default ();
  
  is_active = unity_webapps_interest_manager_get_interest_is_active (manager, interest_id);
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Interest %d %s", interest_id, is_active ? "is active" : "is not active");
  
  g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", is_active, NULL));
  
  return TRUE;
}

static gboolean
on_handle_set_interest_is_active (UnityWebappsGenContext *context,
				  GDBusMethodInvocation *invocation,
				  gint interest_id,
				  gboolean is_active,
				  gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling SetInterestIsActive call (%d)", interest_id);
  
  manager = unity_webapps_interest_manager_get_default ();

  unity_webapps_interest_manager_set_interest_is_active (manager, interest_id,
							 is_active);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}


static gboolean
on_handle_get_interest_window (UnityWebappsGenContext *context,
				  GDBusMethodInvocation *invocation,
				  gint interest_id,
				  gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  guint64 window;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling GetInterestWindow call (%d)", interest_id);
  
  manager = unity_webapps_interest_manager_get_default ();
  
  window = unity_webapps_interest_manager_get_interest_window (manager, interest_id);
  
  g_dbus_method_invocation_return_value (invocation, g_variant_new ("(t)", window, NULL));
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Window for interest (%d) is %lu", interest_id, (unsigned long)window);
  
  return TRUE;
}

static gboolean
on_handle_set_interest_window (UnityWebappsGenContext *context,
				  GDBusMethodInvocation *invocation,
				  gint interest_id,
				  guint64 window,
				  gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling SetInterestWindow call (%d)", interest_id);
  
  manager = unity_webapps_interest_manager_get_default ();

  unity_webapps_interest_manager_set_interest_window (manager, interest_id,
							 window);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}



static gboolean
on_handle_get_interest_location (UnityWebappsGenContext *context,
				 GDBusMethodInvocation *invocation,
				 gint interest_id,
				 gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  const gchar *location;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling GetInterestLocation call (%d)", interest_id);
  
  manager = unity_webapps_interest_manager_get_default ();
  
  location = unity_webapps_interest_manager_get_interest_location (manager, interest_id);
  
  g_dbus_method_invocation_return_value (invocation, g_variant_new ("(s)", location ? location : "", NULL));
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Location for interest (%d) is '%s'", interest_id, location);
  
  return TRUE;
}

static gboolean
on_handle_set_interest_location (UnityWebappsGenContext *context,
				 GDBusMethodInvocation *invocation,
				 gint interest_id,
				 const gchar *location,
				 gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling SetInterestLocation call (%d): %s", interest_id, location);
  
  manager = unity_webapps_interest_manager_get_default ();

  unity_webapps_interest_manager_set_interest_location (manager, interest_id,
							location);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_request_preview (UnityWebappsGenContext *context,
			   GDBusMethodInvocation *invocation,
			   gint interest_id,
			   gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling RequestPreview call");
  
  manager = unity_webapps_interest_manager_get_default();
  
  unity_webapps_context_daemon_emit_preview_requested (session_bus, interest_id);
  
  unity_webapps_interest_manager_add_preview_request (manager, interest_id,
						      invocation);
  
  return TRUE;
}

static gboolean
on_handle_preview_ready (UnityWebappsGenContext *context,
			 GDBusMethodInvocation *invocation,
			 gint interest_id,
			 const gchar *preview_data,
			 gpointer user_data)
{
  UnityWebappsInterestManager *manager;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling PreviewReady call");
  
  manager = unity_webapps_interest_manager_get_default ();
  
  unity_webapps_interest_manager_clear_preview_requests (manager, interest_id, preview_data);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_raise (UnityWebappsGenContext *context,
		    GDBusMethodInvocation *invocation,
		    gpointer user_data)
{
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling Raise call");
  unity_webapps_context_daemon_emit_raise (session_bus, unity_webapps_context_daemon_get_most_recent_interest ());

  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_raise_interest (UnityWebappsGenContext *context,
			  GDBusMethodInvocation *invocation,
			  gint interest_id,
			  gpointer user_data)
{
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling RaiseInterest call (%d)", interest_id);
  unity_webapps_context_daemon_emit_raise (session_bus, interest_id);

  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_add_application_action (UnityWebappsGenContext *context,
				  GDBusMethodInvocation *invocation,
				  const gchar *label,
				  gpointer user_data)
{
  UnityWebappsActionManager *action_manager;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling AddApplicationAction call (%s)", label);
  
  action_manager = unity_webapps_action_manager_get_default ();
  
  unity_webapps_action_manager_add_action (action_manager, label);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_shutdown (UnityWebappsGenContext *context,
		    GDBusMethodInvocation *invocation,
		    gpointer user_data)
{
  UnityWebappsContextDaemon *context_daemon;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Handling Shutdown call");
  
  context_daemon = (UnityWebappsContextDaemon *)user_data;
  
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  g_dbus_connection_flush_sync (session_bus, NULL, NULL);
  
  // TODO: Really should only do this in debug mode?
  unity_webapps_context_daemon_free (context_daemon);
  
  exit (0);
  
  return TRUE;
}

static void
export_object (GDBusConnection *connection,
	       UnityWebappsContextDaemon *daemon)
{
  GError *error;
  
  context_service_interface = unity_webapps_gen_context_skeleton_new ();
  
  g_signal_connect (context_service_interface,
		    "handle-add-icon",
		    G_CALLBACK (on_handle_add_icon),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-get-icon-name",
		    G_CALLBACK (on_handle_get_icon_name),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-add-interest",
		    G_CALLBACK (on_handle_add_interest),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-lost-interest",
		    G_CALLBACK (on_handle_lost_interest),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-list-interests",
		    G_CALLBACK (on_handle_list_interests),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-get-interest-count",
		    G_CALLBACK (on_handle_get_interest_count),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-get-interest-owner",
		    G_CALLBACK (on_handle_get_interest_owner),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-get-view-is-active",
		    G_CALLBACK (on_handle_get_interest_is_active),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-set-view-is-active",
		    G_CALLBACK (on_handle_set_interest_is_active),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-get-view-location",
		    G_CALLBACK (on_handle_get_interest_location),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-set-view-location",
		    G_CALLBACK (on_handle_set_interest_location),
		    daemon);		    
  g_signal_connect (context_service_interface,
		    "handle-request-preview",
		    G_CALLBACK (on_handle_request_preview),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-preview-ready",
		    G_CALLBACK (on_handle_preview_ready),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-get-view-window",
		    G_CALLBACK (on_handle_get_interest_window),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-set-view-window",
		    G_CALLBACK (on_handle_set_interest_window),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-raise",
		    G_CALLBACK (on_handle_raise),
		    daemon);  
  g_signal_connect (context_service_interface,
		    "handle-raise-interest",
		    G_CALLBACK (on_handle_raise_interest),
		    daemon);  
  g_signal_connect (context_service_interface,
		    "handle-add-application-action",
		    G_CALLBACK (on_handle_add_application_action),
		    daemon);
  g_signal_connect (context_service_interface,
		    "handle-shutdown",
		    G_CALLBACK (on_handle_shutdown),
		    daemon);
  
  
  unity_webapps_gen_context_set_view_is_active (context_service_interface, FALSE);
  
  unity_webapps_gen_context_set_name (context_service_interface, info.name);
  unity_webapps_gen_context_set_domain (context_service_interface, info.domain);
  unity_webapps_gen_context_set_desktop_name (context_service_interface, daemon->desktop_file_name);
  
  error = NULL;
  
  g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (context_service_interface),
				    connection,
				    UNITY_WEBAPPS_CONTEXT_PATH,
				    &error);
  
  if (error != NULL)
    {
      g_error ("Error exporting Unity Webapps Context object: %s", error->message);
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Exported Context object");
}

static void
on_manager_became_lonely (UnityWebappsInterestManager *manager,
			  gpointer user_data)
{
  emit_no_interest (session_bus);
}

static void
on_manager_interest_added (UnityWebappsInterestManager *manager,
			   UnityWebappsInterest *interest,
			   gpointer user_data)
{
  emit_interests_changed (session_bus);
}

static void
on_manager_interest_removed (UnityWebappsInterestManager *manager,
			     UnityWebappsInterest *interest,
			     gpointer user_data)
{
  g_queue_remove_all (most_recent_interest_queue, GINT_TO_POINTER (interest->id));

  emit_interests_changed (session_bus);
}

static void
on_manager_active_changed (UnityWebappsInterestManager *manager,
			   gint interest_id,
			   gboolean is_active,
			   gpointer user_data)
{
  unity_webapps_context_daemon_emit_view_is_active_changed (session_bus, interest_id, is_active);
}

static void
on_manager_window_changed (UnityWebappsInterestManager *manager,
			   gint interest_id,
			   guint64 window,
			   gpointer user_data)
{
  unity_webapps_context_daemon_emit_view_window_changed (session_bus, interest_id, window);
}

static void
on_manager_location_changed (UnityWebappsInterestManager *manager,
			     gint interest_id,
			     const gchar *location,
			     gpointer user_data)
{
  unity_webapps_context_daemon_emit_view_location_changed (session_bus, interest_id, location);
}

static void
on_action_manager_action_activated (UnityWebappsActionManager *manager,
				    const gchar *name,
				    gpointer user_data)
{
  unity_webapps_context_daemon_emit_action_invoked (session_bus, name);
}

UnityWebappsContextDaemon *
unity_webapps_context_daemon_new ()
{
  UnityWebappsContextDaemon *daemon;
  UnityWebappsInterestManager *manager;
  UnityWebappsActionManager *action_manager;
  gchar *t, *icon_path;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Constructing new UnityWebappsContextDaemon object");
  
  daemon = g_malloc0 (sizeof (UnityWebappsContextDaemon));
  
  //  daemon->icon_theme_dir = unity_webapps_icon_theme_get_theme_dir ();
  daemon->icon_theme_dir = NULL;
  
  daemon->interests = 0;
  
  daemon->interest_vanished_ids = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  daemon->name_interest_counts = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  
  if ((info.icon_url != NULL) && (info.icon_is_themed == FALSE))
    {
      daemon->icon_file_name = save_icon_file (info.name, info.domain, info.icon_url);
    }
  else 
    {
      daemon->icon_file_name = NULL;
    }

  daemon->desktop_file_name = get_desktop_file_name (daemon);
  icon_path = unity_webapps_icon_theme_get_application_icon_path (daemon->icon_file_name);
  
  daemon->notification_context = unity_webapps_notification_context_new (session_bus, info.name, icon_path);
  t = get_canonical_full_name(info.name, info.domain);
  daemon->indicator_context = unity_webapps_indicator_context_new (session_bus, daemon->desktop_file_name, t);
  daemon->music_player_context = unity_webapps_music_player_context_new (session_bus, daemon->desktop_file_name, t, info.name);
  daemon->launcher_context = unity_webapps_launcher_context_new (session_bus, daemon->desktop_file_name);
  
  
  manager = unity_webapps_interest_manager_get_default ();
  
  g_signal_connect (manager, "became-lonely",
		    G_CALLBACK (on_manager_became_lonely),
		    daemon);
  g_signal_connect (manager, "interest-added",
		    G_CALLBACK (on_manager_interest_added),
		    daemon);
  g_signal_connect (manager, "interest-removed",
		    G_CALLBACK (on_manager_interest_removed),
		    daemon);
  g_signal_connect (manager, "active-changed",
		    G_CALLBACK (on_manager_active_changed),
		    daemon);
  g_signal_connect (manager, "location-changed",
		    G_CALLBACK (on_manager_location_changed),
		    daemon);
  g_signal_connect (manager, "window-changed",
		    G_CALLBACK (on_manager_window_changed),
		    daemon);

  action_manager = unity_webapps_action_manager_get_default ();
  
  g_signal_connect (action_manager, "action-invoked",
		    G_CALLBACK (on_action_manager_action_activated),
		    daemon);
  
  most_recent_interest_queue = g_queue_new();
  

  
  g_free (t);
  g_free (icon_path);

  return daemon;
}

/*
 * TODO: This probably leaks...but in the current implementation this means we are shutting down.
 * It would be useful to free everything just to aid memory debuggers.
 */
void
unity_webapps_context_daemon_free (UnityWebappsContextDaemon *daemon)
{
  // TODO: Can we come back after this?
  g_hash_table_destroy (daemon->interest_vanished_ids);
  g_hash_table_destroy (daemon->name_interest_counts);
  
  g_queue_free (most_recent_interest_queue);
  
  unity_webapps_notification_context_free (daemon->notification_context);
  unity_webapps_indicator_context_free (daemon->indicator_context);
  unity_webapps_music_player_context_free (daemon->music_player_context);
  unity_webapps_launcher_context_free (daemon->launcher_context);
  
  
  g_free (daemon->desktop_file_name);
  g_free (daemon->icon_file_name);
  g_free (daemon->icon_theme_dir);

  g_free (daemon);
}

void
unity_webapps_context_daemon_emit_raise (GDBusConnection *connection, gint interest_id)
{
  GError *error;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting Raise signal to interest: %d", interest_id);
  
  if ((interest_id < 0) && (unity_webapps_context_daemon_get_most_recent_interest() > 0))
    {
      interest_id = unity_webapps_context_daemon_get_most_recent_interest();
    }
  
  error = NULL;
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "RaiseRequested",
				 g_variant_new ("(i)",
						interest_id,
						NULL),
				 &error);
  
  if (error != NULL)
    {
      g_warning ("Error emitting raise signal in context: %s", error->message);
      
      g_error_free (error);

      return;
    }

}

void 
unity_webapps_context_daemon_emit_view_is_active_changed (GDBusConnection *connection, gint interest_id, gboolean is_active)
{
  GError *error;

  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting ViewIsActiveChanged signal (interest %d %s)", interest_id,
		      is_active ? "is active" : "is not active");
  
  error = NULL;
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "ViewIsActiveChanged",
				 g_variant_new ("(ib)",
						interest_id,
						is_active,
						NULL),
				 &error);
  
  if (error != NULL)
    {
      g_warning ("Error emitting ViewIsActiveChanged signal in context: %s", error->message);
      
      g_error_free (error);

      return;
    }
}


void 
unity_webapps_context_daemon_emit_view_window_changed (GDBusConnection *connection, gint interest_id, guint64 window)
{
  GError *error;
    
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting ViewWindowChanged signal (interest %d has window %lu)", interest_id, (unsigned long) window);
  
  error = NULL;
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "ViewWindowChanged",
				 g_variant_new ("(it)",
						interest_id,
						window,
						NULL),
				 &error);
  
  if (error != NULL)
    {
      g_warning ("Error emitting ViewWindowChanged signal in context: %s", error->message);
      
      g_error_free (error);

      return;
    }
}

void 
unity_webapps_context_daemon_emit_view_location_changed (GDBusConnection *connection, gint interest_id, const gchar *location)
{
  GError *error;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting ViewLocationChanged signal (interest %d navigated to %s)",
		      interest_id, location);
  
  error = NULL;
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "ViewLocationChanged",
				 g_variant_new ("(is)",
						interest_id,
						location,
						NULL),
				 &error);
  
  if (error != NULL)
    {
      g_warning ("Error emitting ViewLocationChanged signal in context: %s", error->message);
      
      g_error_free (error);

      return;
    }
}

void 
unity_webapps_context_daemon_emit_preview_requested (GDBusConnection *connection, gint interest_id)
{
  GError *error;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting PreviewRequested signal to interest %d", interest_id);
  
  error = NULL;
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "PreviewRequested",
				 g_variant_new ("(i)",
						interest_id,
						NULL),
				 &error);
  
  if (error != NULL)
    {
      g_warning ("Error emitting PreviewRequested signal in context: %s", error->message);
      
      g_error_free (error);

      return;
    }
}


void 
unity_webapps_context_daemon_emit_action_invoked (GDBusConnection *connection, const gchar *name)
{
  GError *error;
  
  UNITY_WEBAPPS_NOTE (CONTEXT, "Emitting ApplicationActionInvoked signal (%s)", name);
  
  error = NULL;
  g_dbus_connection_emit_signal (connection,
				 NULL,
				 UNITY_WEBAPPS_CONTEXT_PATH,
				 UNITY_WEBAPPS_CONTEXT_IFACE,
				 "ApplicationActionInvoked",
				 g_variant_new ("(s)",
						name,
						NULL),
				 &error);
  
  if (error != NULL)
    {
      g_warning ("Error emitting ApplicationActionInvoked signal in context: %s", error->message);
      
      g_error_free (error);

      return;
    }
}

gint
unity_webapps_context_daemon_get_most_recent_interest ()
{
  if (g_queue_get_length (most_recent_interest_queue) == 0)
    {
      return -1;
    }
  else
    {
      return GPOINTER_TO_INT (g_queue_pop_tail (most_recent_interest_queue));
    }
}

/* 
 * Update the most recent client. This is called when the implementation contexts (launcher, indicator, notification,
 * music player) receive message. The "Raise" signal will only be dispatched to this most recent client.
 */
void 
unity_webapps_context_daemon_set_most_recent_interest (gint interest)
{
  UNITY_WEBAPPS_NOTE (CONTEXT, "Updating most recent interest: %d", interest);
  g_queue_push_tail (most_recent_interest_queue, GINT_TO_POINTER (interest));
}
