/*
 * unity-webapps-dirs.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <sys/stat.h>
#include <errno.h>

#include <glib/gstdio.h>

#include "unity-webapps-dirs.h"

/*
 * Get the unity webapps user storage directory (and ensure it exists).
 * Currently this is $HOME/.local/share/unity-webapps
 */

gchar *
unity_webapps_dirs_get_user_dir ()
{
  const gchar *homedir;
  gchar *userdir;
  
  homedir = g_getenv ("HOME");
  if (!homedir)
    homedir = g_get_home_dir ();
  
  userdir = g_build_filename(g_get_user_data_dir(), "unity-webapps", NULL);
  
  if (g_file_test (userdir, G_FILE_TEST_EXISTS) == FALSE)
    {
      int status;

      status = g_mkdir (userdir, 0700);
      
      if (status < -1)
	{
	  g_critical ("Error making Unity Webapps user dir (%s): %s", userdir, g_strerror(errno));
	  
	  g_free (userdir);
	  
	  return NULL;
	}
    }
  
  return userdir;  
}

/*
 * Get the unity webapps icon storage directory and ensure it exists, 
 * currently this is a folder "icons" inside the user storage directory. */
gchar *
unity_webapps_dirs_get_icon_dir () 
{
  gchar *icondir;
  const gchar *envdir;
  
  envdir = g_getenv("UNITY_WEBAPPS_CONTEXT_ICON_DIR");
  if (envdir != NULL)
    {
      return g_strdup (envdir);
    }
  
  icondir = g_build_filename (g_get_home_dir(), ".icons", NULL);
  
  if (g_file_test (icondir, G_FILE_TEST_EXISTS) == FALSE)
    {
      int status; 

      status = g_mkdir (icondir, 0700);
      
      if (status < -1)
	{
	  g_critical ("Error making Unity Webapps icon dir (%s): %s", icondir, g_strerror(errno));
	  
	  g_free (icondir);
	  
	  return NULL;
	}
    }
  
  return icondir;
}

gchar *
unity_webapps_dirs_get_icon_theme_dir () 
{
  gchar *icondir;
  const gchar *envdir;
  
  envdir = g_getenv("UNITY_WEBAPPS_CONTEXT_ICON_THEME_DIR");
  if (envdir != NULL)
    {
      return g_strdup (envdir);
    }
  
  icondir = g_build_filename (g_get_home_dir(), ".local", "share", "icons", NULL);
  
  if (g_file_test (icondir, G_FILE_TEST_EXISTS) == FALSE)
    {
      int status; 

      status = g_mkdir (icondir, 0700);
      
      if (status < -1)
	{
	  g_critical ("Error making Unity Webapps icon theme dir (%s): %s", icondir, g_strerror(errno));
	  
	  g_free (icondir);
	  
	  return NULL;
	}
    }
  
  g_free (icondir);

  icondir = g_build_filename (g_get_home_dir(), ".local", "share", "icons", "unity-webapps", NULL);
  
  if (g_file_test (icondir, G_FILE_TEST_EXISTS) == FALSE)
    {
      int status; 

      status = g_mkdir (icondir, 0700);
      
      if (status < -1)
	{
	  g_critical ("Error making Unity Webapps icon theme dir (%s): %s", icondir, g_strerror(errno));
	  
	  g_free (icondir);
	  
	  return NULL;
	}
    }
  
  return icondir;
}

/* 
 * Get the directory to store desktop files in
 */
gchar *
unity_webapps_dirs_get_application_dir () 
{
  gchar *applicationdir;
  const gchar *envdir;

  envdir = g_getenv("UNITY_WEBAPPS_CONTEXT_APPLICATION_DIR");
  if (envdir != NULL)
    {
      return g_strdup (envdir);
    }
 
  /* 
   * This is mostly useful for testing, i.e. we want to write the .desktop file to a temporary directory
   * and verify its contents.
   * TODO: We should have a test that does this!
   */
  applicationdir = g_build_filename (g_get_user_data_dir(), "applications", NULL);
  
  if (g_file_test (applicationdir, G_FILE_TEST_EXISTS) == FALSE)
    {
      int status; 

      status = g_mkdir (applicationdir, 0700);
      
      if (status < -1)
	{
	  g_critical ("Error making Unity Webapps application dir (%s): %s", applicationdir, g_strerror(errno));
	  
	  g_free (applicationdir);
	  
	  return NULL;
	}
    }
  
  
  return applicationdir;
}



gchar *
unity_webapps_dirs_get_resource_dir () 
{
  gchar *resourcedir;
  gchar *userdir;
  const gchar *envdir;

  envdir = g_getenv("UNITY_WEBAPPS_CONTEXT_APPLICATION_DIR");
  if (envdir != NULL)
    {
      return g_strdup (envdir);
    }
 
  userdir = unity_webapps_dirs_get_user_dir ();
  resourcedir = g_build_filename (userdir, "resources", NULL);
  g_free (userdir);
  
  if (g_file_test (resourcedir, G_FILE_TEST_EXISTS) == FALSE)
    {
      int status; 

      status = g_mkdir (resourcedir, 0700);
      
      if (status < -1)
	{
	  g_critical ("Error making Unity Webapps resource dir (%s): %s", resourcedir, g_strerror(errno));
	  
	  g_free (resourcedir);
	  
	  return NULL;
	}
    }
  
  
  return resourcedir;
}


