/*
 * unity-webapps-indicator-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <libindicate-gtk/indicator.h>
#include <gtk/gtk.h>

#include "unity-webapps-indicator-context.h"
#include "unity-webapps-resource-cache.h"

#include "../unity-webapps-dbus-defs.h"

#include "unity-webapps-dirs.h"
#include "../unity-webapps-debug.h"

#include "unity-webapps-telepathy-presence-manager.h"

#include "config.h"

static gboolean menuitem_has_child_with_label (DbusmenuMenuitem *item,
					       const gchar *label);

static void on_menuitem_activated (DbusmenuMenuitem *item,
				   guint timestamp,
				   gpointer user_data);

static void real_add_action (UnityWebappsIndicatorContext *indicator_context,
			     const gchar *label);

static void real_clear_indicator (UnityWebappsIndicatorContext *indicator_context,
				  const gchar *name);

static void real_set_property_icon (UnityWebappsIndicatorContext *indicator_context,
				    const gchar *name,
				    const gchar *property,
				    const gchar *icon_url);

static void real_set_property (UnityWebappsIndicatorContext *indicator_context,
			       const gchar *name,
			       const gchar *property,
			       const gchar *value);

static void real_show_indicator (UnityWebappsIndicatorContext *indicator_context,
				 const gchar *name);

static void server_displayed_callback (IndicateServer *server,
				       guint timestamp,
				       gpointer user_data);


static void create_indicate_server (UnityWebappsIndicatorContext *indicator_context);

static IndicateServer *unity_webapps_indicator_context_get_server (UnityWebappsIndicatorContext *indicator_context);

#define UNITY_WEBAPPS_INDICATOR_MAX_ACTIONS 4
#define UNITY_WEBAPPS_INDICATOR_MAX_INDICATORS 5


/* 
 * If anything wants to use indicator_context->indicate_server you have to get
 * it through this. This prevents us from making anything on DBus until it is
 * actually used
*/
static IndicateServer *
unity_webapps_indicator_context_get_server (UnityWebappsIndicatorContext *indicator_context)
{
  if (indicator_context->indicate_server)
    return indicator_context->indicate_server;
  else
    {
      create_indicate_server (indicator_context);
      return indicator_context->indicate_server;
    }
  
}

/* 
 * Search if a Dbusmenu item has a child with a certain label, we use this
 * to avoid inserting duplicate actions
 */
static gboolean
menuitem_has_child_with_label (DbusmenuMenuitem *item,
			       const gchar *label)
{
  const GList *children;
  const GList *walk;
  
  children = dbusmenu_menuitem_get_children (item);
  
  for (walk = children;
       walk != NULL;
       walk = walk->next)
    {
      DbusmenuMenuitem *child;
      const gchar *child_label;
      
      child = (DbusmenuMenuitem *)walk->data;
      child_label = dbusmenu_menuitem_property_get (child, "label");
      
      if (g_strcmp0 (child_label, label) == 0)
	{
	  return TRUE;
	}
      
    }
  return FALSE;
}

static void
emit_action_invoked_signal (UnityWebappsIndicatorContext *indicator_context,
			    const gchar *action)
{
  GError *error;

  UNITY_WEBAPPS_NOTE (INDICATOR, "Emitting ActionInvoked signal (%s)", action);

  error = NULL;
  g_dbus_connection_emit_signal (indicator_context->connection,
				 NULL,
				 UNITY_WEBAPPS_INDICATOR_PATH,
				 UNITY_WEBAPPS_INDICATOR_IFACE,
				 "ActionInvoked",
				 g_variant_new ("(s)", (gchar *)action, NULL),
				 &error);
  if (error != NULL)
    {
      g_warning ("Error emitting ActionInvoked signal (from indicator) in indicator context: %s", error->message);
      
      g_error_free (error);
    }


  
}

static void
on_indicator_user_display (IndicateIndicator *indicator,
			   guint timestamp,
			   gpointer user_data)
{
  UnityWebappsIndicatorContext *indicator_context;
  GHashTableIter iter;
  gpointer key, value;
  gboolean found;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Handling Indicator user-display signal");
  
  indicator_context = (UnityWebappsIndicatorContext *)user_data;
  
  g_hash_table_iter_init (&iter, indicator_context->indicators_by_name);
  
  found = FALSE;
  
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      if (value == indicator)
	{
	  found = TRUE;
	  break;
	}
    }
  
  if (found == FALSE)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Got user-display signal for an indicator we don't know about (%p)", indicator);
      return;
    }
  
  emit_action_invoked_signal (indicator_context,
			      (gchar *)key);
}

/*
 * When an action menuitem is activated. We need to emit our action
 * signal.
 */
static void
on_menuitem_activated (DbusmenuMenuitem *item,
		       guint timestamp,
		       gpointer user_data)
{
  UnityWebappsIndicatorContext *indicator_context;
  const gchar *label;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Handling Menuitem activated signal");
  
  indicator_context = (UnityWebappsIndicatorContext *)user_data;

  unity_webapps_context_daemon_emit_raise (indicator_context->connection, -1);
  
  label = dbusmenu_menuitem_property_get (item, DBUSMENU_MENUITEM_PROP_LABEL);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Menuitem label is: %s", label);
  
  emit_action_invoked_signal (indicator_context, label);
}

/*
 * Add an action with a label.
 * an action roughly corresponds to a DBusMenu item, the child of a menu set on the indicate server.
 * When the action is activated, the ActionInvoked callback will be emitted over DBus with the action label.
 */
static void
real_add_action (UnityWebappsIndicatorContext *indicator_context,
		 const gchar *label)
{
  DbusmenuMenuitem *action_item;
  
  indicate_server_show (unity_webapps_indicator_context_get_server (indicator_context));

  if (indicator_context->num_actions > UNITY_WEBAPPS_INDICATOR_MAX_ACTIONS)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Tried to add action (%s), but number of actions is greater than UNITY_WEBAPPS_INDICATOR_MAX_ACTIONS (%d)",
			  label, UNITY_WEBAPPS_INDICATOR_MAX_ACTIONS);
      return;
    }
  
  if (menuitem_has_child_with_label (indicator_context->menu_item, label))
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Tried to add action (%s), but it already exists. Continuing", label);
      return;
    }
  
  unity_webapps_indicator_context_get_server (indicator_context);
  
  indicator_context->num_actions++;
  
  action_item = dbusmenu_menuitem_new ();
  
  dbusmenu_menuitem_property_set (action_item, DBUSMENU_MENUITEM_PROP_LABEL, label);
  dbusmenu_menuitem_property_set_bool (action_item, DBUSMENU_MENUITEM_PROP_VISIBLE, TRUE);
  
  g_signal_connect (action_item, "item-activated", G_CALLBACK(on_menuitem_activated), indicator_context);
  
  dbusmenu_menuitem_child_append (indicator_context->menu_item, action_item);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Added indicator action (%s)", label);
}

/* Delete an indicator by name */
static void
real_clear_indicator (UnityWebappsIndicatorContext *indicator_context,
		      const gchar *name)
{
  IndicateIndicator *indicator;

  indicator = g_hash_table_lookup (indicator_context->indicators_by_name,
				   name);
  
  if (indicator == NULL)
    return;
  
  indicate_indicator_hide (indicator);
  
  g_hash_table_remove (indicator_context->indicators_by_name, name);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Cleared and freed indicator: %s", name);
}

static void
real_set_property_icon (UnityWebappsIndicatorContext *indicator_context,
			const gchar *name,
			const gchar *property,
			const gchar *icon_url)
{
  IndicateIndicator *indicator;
  GFile *icon_file;
  GFileInputStream *input_stream;
  GdkPixbuf *pb;
  gchar *icon_path;
  GError *error;
  
  indicator = g_hash_table_lookup (indicator_context->indicators_by_name, name);
  
  if (indicator == NULL)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Handling SetPropertyIcon for %s but there is no indicator by that name", name);
      return;
    }
  
  
  icon_path = unity_webapps_resource_cache_lookup_uri (icon_url);
  
  if (icon_path == NULL)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Resource cache returned NULL for URL (%s)", icon_url);
      return;
    }
  
  icon_file = g_file_new_for_path (icon_path);
  
  if (icon_file == NULL)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Failed to construct file from icon path: %s", icon_path);
      return;
    }
  
  error = NULL;
  input_stream = g_file_read (icon_file, NULL, &error);
  
  if (error != NULL)
    {
      g_warning ("Error downloading icon url in indicator context: %s",
		 icon_url);

      g_error_free (error);
      
      g_object_unref (G_OBJECT (icon_file));
      
      return;
    }

  pb = gdk_pixbuf_new_from_stream (G_INPUT_STREAM (input_stream), NULL, &error);
  
  if (error != NULL)
    {
      g_warning ("Could not create pixbuf from image stream (%s): %s", icon_url, error->message);
      g_error_free (error);
      g_object_unref (G_OBJECT (icon_file));
      g_object_unref (G_OBJECT (input_stream));
    }
  
  indicate_gtk_indicator_set_property_icon (indicator, property, pb);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Set icon property of %s", name);
  
  g_object_unref (pb);
  g_object_unref (G_OBJECT (icon_file));
  g_object_unref (G_OBJECT (input_stream));
}

static void
real_set_property (UnityWebappsIndicatorContext *indicator_context,
		   const gchar *name,
		   const gchar *property,
		   const gchar *value)
{
  IndicateIndicator *indicator;
  
  indicator = g_hash_table_lookup (indicator_context->indicators_by_name, name);
  
  if (indicator == NULL)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Tried to set %s property of indicator %s but no such indicator exists",
			  property, name);
      return;
    }
  
  indicate_indicator_set_property (indicator, property, value);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Set property %s of indicator %s", property, name);
}

static void
real_show_indicator (UnityWebappsIndicatorContext *indicator_context,
		     const gchar *name)
{
  IndicateIndicator *indicator;
  IndicateServer *server;
  
  if (g_hash_table_size (indicator_context->indicators_by_name) > UNITY_WEBAPPS_INDICATOR_MAX_INDICATORS)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Tried to show indicator %s but number of indicators is greater than UNITY_WEBAPPS_INDICATOR_MAX_INDICATORS (%d)",
			  name, UNITY_WEBAPPS_INDICATOR_MAX_INDICATORS);
      return;
    }
  
  if ((indicator = g_hash_table_lookup (indicator_context->indicators_by_name, name)) != NULL)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Showing indicator %s found existing indicator, ensuring it is visible", name);

      indicate_indicator_show (indicator);
      return;
    }
  
  server = unity_webapps_indicator_context_get_server (indicator_context);
  
  indicator = indicate_indicator_new_with_server (server);
  indicate_indicator_set_property (indicator, "name", name);
  
  indicate_indicator_show (indicator);
  
  g_signal_connect (indicator, "user-display",
		    G_CALLBACK (on_indicator_user_display), 
		    indicator_context);
  
  g_hash_table_insert (indicator_context->indicators_by_name, g_strdup (name), indicator);
}

static void
server_displayed_callback (IndicateServer *server,
			   guint timestamp,
			   gpointer user_data)
{
  UnityWebappsIndicatorContext *indicator_context;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Got server-displayed callback");
  
  indicator_context = (UnityWebappsIndicatorContext *)user_data;

  
  unity_webapps_context_daemon_emit_raise (indicator_context->connection, -1);
				
}

static void
create_indicate_server (UnityWebappsIndicatorContext *indicator_context)
{
  gchar *path, *menu_dbus_path, *applications_dir;

  indicate_server_ref_default (); // HACK (from racarr) libindicate race condition bug.
  /*
   * If we don't leak the default server, the following race condition happens:
   * 1. We create an indicator
   * 2. The indicator refs the default server, creating it (unless we have already leaked it...)
   * 3. The default server tries to open a DBus connection asynchronously.
   * 4. libindicate sets the indicator server to the webapps server, unreffing the default server.
   * 5. The default server is destroyed.
   * 6. The bus_get_cb from the default server is called, and KABOOM!
   */

  indicator_context->indicate_server = g_object_new (INDICATE_TYPE_SERVER, NULL); //"path", "/com/canonical/Unity/WebApps/Context/IndicateServer", NULL);
  
  g_signal_connect (indicator_context->indicate_server,
		    "server-display", G_CALLBACK (server_displayed_callback),
		    indicator_context);
  
  // TODO: Fixme
  indicate_server_set_type (indicator_context->indicate_server, "message.im");
  
  applications_dir = unity_webapps_dirs_get_application_dir ();

  path = g_build_filename (applications_dir, indicator_context->desktop_name, NULL);
  indicate_server_set_desktop_file (indicator_context->indicate_server, path);
  
  menu_dbus_path = g_strdup_printf("/com/canonical/Unity/WebApps/%s/indicate/menu", indicator_context->canonical_name);
  
  indicator_context->menu_item = dbusmenu_menuitem_new ();
  indicator_context->menu_server = dbusmenu_server_new (menu_dbus_path);
  
  dbusmenu_server_set_root (indicator_context->menu_server, indicator_context->menu_item);
  
  indicate_server_set_menu (indicator_context->indicate_server, indicator_context->menu_server);
  
  
  g_free (menu_dbus_path);
  g_free (path);
  g_free (applications_dir);
}

static gboolean
on_handle_show_indicator (UnityWebappsGenIndicator *indicator,
			  GDBusMethodInvocation *invocation,
			  const gchar *name,
			  gpointer user_data)
{
  UnityWebappsIndicatorContext *context;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Handling ShowIndicator call: %s", name);
  
  context = (UnityWebappsIndicatorContext *)user_data;
  
  real_show_indicator (context, name);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_clear_indicator (UnityWebappsGenIndicator *indicator,
			  GDBusMethodInvocation *invocation,
			  const gchar *name,
			  gpointer user_data)
{
  UnityWebappsIndicatorContext *context;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Handling ClearIndicator call: %s", name);
  
  context = (UnityWebappsIndicatorContext *)user_data;
  
  real_clear_indicator (context, name);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_set_property (UnityWebappsGenIndicator *indicator,
			GDBusMethodInvocation *invocation,
			const gchar *name,
			const gchar *property,
			const gchar *value,
			gpointer user_data)
{
  UnityWebappsIndicatorContext *context;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Handling SetProperty call for %s: %s->%s",
		      name, property, value);
  
  context = (UnityWebappsIndicatorContext *)user_data;
  
  real_set_property (context, name, property, value);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}


static gboolean
on_handle_set_property_icon (UnityWebappsGenIndicator *indicator,
			     GDBusMethodInvocation *invocation,
			     const gchar *name,
			     const gchar *property,
			     const gchar *value,
			     gpointer user_data)
{
  UnityWebappsIndicatorContext *context;

  UNITY_WEBAPPS_NOTE (INDICATOR, "Handling SetPropertyIcon call for %s: %s->%s",
		      name, property, value);
  
  context = (UnityWebappsIndicatorContext *)user_data;
  
  real_set_property_icon (context, name, property, value);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_add_action (UnityWebappsGenIndicator *indicator,
		      GDBusMethodInvocation *invocation,
		      const gchar *label,
		      gpointer user_data)
{
  UnityWebappsIndicatorContext *context;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Handling AddAction call, label: %s", label);
  
  context = (UnityWebappsIndicatorContext *)user_data;
  
  real_add_action (context, label);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}



static void
export_object (GDBusConnection *connection,
	       UnityWebappsIndicatorContext *indicator_context)
{
  GError *error;
  
  indicator_context->indicator_service_interface = unity_webapps_gen_indicator_skeleton_new ();
  
  g_signal_connect (indicator_context->indicator_service_interface,
		    "handle-show-indicator",
		    G_CALLBACK (on_handle_show_indicator),
		    indicator_context);
  g_signal_connect (indicator_context->indicator_service_interface,
		    "handle-clear-indicator",
		    G_CALLBACK (on_handle_clear_indicator),
		    indicator_context);
  g_signal_connect (indicator_context->indicator_service_interface,
		    "handle-set-property",
		    G_CALLBACK (on_handle_set_property),
		    indicator_context);
  g_signal_connect (indicator_context->indicator_service_interface,
		    "handle-set-property-icon",
		    G_CALLBACK (on_handle_set_property_icon),
		    indicator_context);
  g_signal_connect (indicator_context->indicator_service_interface,
		    "handle-add-action",
		    G_CALLBACK (on_handle_add_action),
		    indicator_context);
  
  error = NULL;
  
  g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (indicator_context->indicator_service_interface),
				    connection,
				    UNITY_WEBAPPS_INDICATOR_PATH,
				    &error);
  
  if (error != NULL)
    {
      g_error ("Error exporting Unity Webapps Indicator object: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Exported Indicator object");

}

static void
on_presence_changed (UnityWebappsPresenceManager *presence_manager,
		     gchar *status,
		     gchar *message,
		     gpointer user_data)
{
  UnityWebappsIndicatorContext *indicator_context;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Most available presence changed to %s", status);
  
  indicator_context = (UnityWebappsIndicatorContext *)user_data;
  
  unity_webapps_gen_indicator_set_presence (indicator_context->indicator_service_interface, status);
}

static void
create_presence_manager (UnityWebappsIndicatorContext *indicator_context)
{
  indicator_context->presence_manager = unity_webapps_telepathy_presence_manager_new ();
  
  g_signal_connect (indicator_context->presence_manager, "presence-changed",
		    G_CALLBACK (on_presence_changed), indicator_context);
}

UnityWebappsIndicatorContext *
unity_webapps_indicator_context_new (GDBusConnection *connection, const gchar *desktop_name, const gchar *canonical_name)
{
  UnityWebappsIndicatorContext *indicator_context;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Creating new UnityWebappsIndicatorContext object");

  indicator_context = g_malloc0 (sizeof (UnityWebappsIndicatorContext));
  
  indicator_context->connection = g_object_ref (G_OBJECT (connection));
  indicator_context->canonical_name = g_strdup (canonical_name);
  indicator_context->desktop_name = g_strdup (desktop_name);
  
  indicator_context->indicators_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
  
  indicator_context->num_actions = 0;
  
  export_object (connection, indicator_context);
  
  //  create_indicate_server (indicator_context);
  
  create_presence_manager (indicator_context);
  
  return indicator_context;
}

void
unity_webapps_indicator_context_free (UnityWebappsIndicatorContext *indicator_context)
{
  UNITY_WEBAPPS_NOTE (INDICATOR, "Finalizing UnityWebappsIndicatorContext object");
  
  g_hash_table_destroy (indicator_context->indicators_by_name);
  g_object_unref (G_OBJECT (indicator_context->indicator_service_interface));
  
  g_free (indicator_context->canonical_name);
  g_free (indicator_context->desktop_name);

  g_object_unref (G_OBJECT (indicator_context->connection));
  g_object_unref (G_OBJECT (indicator_context->presence_manager));
  
  if (indicator_context->indicate_server != NULL)
    g_object_unref (G_OBJECT (indicator_context->indicate_server));

  g_free (indicator_context);
}
