
/*
 * unity-webapps-launcher-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include "unity-webapps-launcher-context.h"

#include "../unity-webapps-dbus-defs.h"

#include "../unity-webapps-debug.h"

#define UNITY_WEBAPPS_LAUNCHER_MAX_ACTIONS 6

static gboolean
menuitem_has_child_with_label (DbusmenuMenuitem *item,
			       const gchar *label)
{
  const GList *children;
  const GList *walk;
  
  if (item == NULL)
    {
      return FALSE;
    }

  
  children = dbusmenu_menuitem_get_children (item);
  
  for (walk = children;
       walk != NULL;
       walk = walk->next)
    {
      DbusmenuMenuitem *child;
      const gchar *child_label;
      
      child = (DbusmenuMenuitem *)walk->data;
      child_label = dbusmenu_menuitem_property_get (child, "label");
      
      if (g_strcmp0 (child_label, label) == 0)
	{
	  return TRUE;
	}
      
    }
  return FALSE;
}

static void
emit_action_invoked_signal (UnityWebappsLauncherContext *launcher_context,
			    const gchar *label)
{
  GError *error;

  UNITY_WEBAPPS_NOTE (LAUNCHER, "Emitting ActionInvoked signal (%s)", label);

  error = NULL;
  
  g_dbus_connection_emit_signal (launcher_context->connection,
				 NULL,
				 UNITY_WEBAPPS_LAUNCHER_PATH,
				 UNITY_WEBAPPS_LAUNCHER_IFACE,
				 "ActionInvoked",
				 g_variant_new("(s)", label, NULL),
				 &error);
  
  if (error != NULL)
    {
      g_warning ("Error emitting ActionInvoked signal (from menuitem) in launcher context: %s", error->message);
      g_error_free (error);
    }

}

static void
on_menuitem_activated (DbusmenuMenuitem *item,
		       guint timestamp,
		       gpointer user_data)
{
  UnityWebappsLauncherContext *launcher_context;
  const gchar *label;
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Got item-activated signal for menuitem");
  
  launcher_context = (UnityWebappsLauncherContext *)user_data;
  
  label = dbusmenu_menuitem_property_get (item, DBUSMENU_MENUITEM_PROP_LABEL);
  
  if (label == NULL)
    {
      UNITY_WEBAPPS_NOTE (LAUNCHER, "Got item-activated signal for menuitem (%p) but label is NULL", item);
      
      return;
    }
  
  emit_action_invoked_signal (launcher_context, label);
  
}



static DbusmenuMenuitem *
launcher_context_get_menuitem (UnityWebappsLauncherContext *launcher_context)
{
  if (launcher_context->menu_item == NULL)
    {
      launcher_context->menu_item = dbusmenu_menuitem_new ();
      unity_launcher_entry_set_quicklist (launcher_context->launcher_entry,
					  launcher_context->menu_item);
    }
  
  return launcher_context->menu_item;
}

static void
real_add_action (UnityWebappsLauncherContext *launcher_context,
		 const gchar *label)
{
  DbusmenuMenuitem *action_item, *root_item;
  
  
  if (launcher_context->num_actions > UNITY_WEBAPPS_LAUNCHER_MAX_ACTIONS)
    {
      return;
    }
  
  if (menuitem_has_child_with_label (launcher_context->menu_item, label))
    {
      return;
    }
  
  root_item = launcher_context_get_menuitem (launcher_context);
  
  launcher_context->num_actions++;
  
  action_item = dbusmenu_menuitem_new ();
  
  dbusmenu_menuitem_property_set (action_item, DBUSMENU_MENUITEM_PROP_LABEL, label);
  dbusmenu_menuitem_property_set_bool (action_item, DBUSMENU_MENUITEM_PROP_VISIBLE, TRUE);
  
  g_signal_connect (action_item, "item-activated", G_CALLBACK(on_menuitem_activated), launcher_context);
  
  dbusmenu_menuitem_child_append (root_item, action_item);
}




static void
real_set_count (UnityWebappsLauncherContext *context,
		gint count)
{
  unity_launcher_entry_set_count (context->launcher_entry, count);
  unity_launcher_entry_set_count_visible (context->launcher_entry, TRUE);
}

static void
real_clear_count (UnityWebappsLauncherContext *context)
{
  unity_launcher_entry_set_count_visible (context->launcher_entry, FALSE);
}
static void
real_set_progress (UnityWebappsLauncherContext *context,
		   gdouble progress)
{
  unity_launcher_entry_set_progress (context->launcher_entry, progress);
  unity_launcher_entry_set_progress_visible (context->launcher_entry, TRUE);
}

static void
real_clear_progress (UnityWebappsLauncherContext *context)
{
  unity_launcher_entry_set_progress_visible (context->launcher_entry, FALSE);
}

static void
real_set_urgent (UnityWebappsLauncherContext *context)
{
  unity_launcher_entry_set_urgent (context->launcher_entry, TRUE);
}

static gboolean
on_handle_set_count (UnityWebappsGenLauncher *launcher,
		     GDBusMethodInvocation *invocation,
		     gint count,
		     gpointer user_data)
{
  UnityWebappsLauncherContext *context;
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Handling SetCount call, count: %d", count);
  
  context = (UnityWebappsLauncherContext *)user_data;

  real_set_count (context, count);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_clear_count (UnityWebappsGenLauncher *launcher,
		       GDBusMethodInvocation *invocation,
		       gpointer user_data)
{
  UnityWebappsLauncherContext *context;
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Handling ClearCount call");
  
  context = (UnityWebappsLauncherContext *)user_data;

  real_clear_count (context);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_set_progress (UnityWebappsGenLauncher *launcher,
			GDBusMethodInvocation *invocation,
			gdouble progress,
			gpointer user_data)
{
  UnityWebappsLauncherContext *context;
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Handling SetProgress call, progress: %f", progress);
  
  context = (UnityWebappsLauncherContext *)user_data;

  real_set_progress (context, progress);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_clear_progress (UnityWebappsGenLauncher *launcher,
			  GDBusMethodInvocation *invocation,
			  gpointer user_data)
{
  UnityWebappsLauncherContext *context;
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Handling ClearProgress call");
  
  context = (UnityWebappsLauncherContext *)user_data;

  real_clear_progress (context);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_add_action (UnityWebappsGenLauncher *launcher,
		      GDBusMethodInvocation *invocation,
		      const gchar *label,
		      gpointer user_data)
{
  UnityWebappsLauncherContext *context;
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Handling AddAction call, action: %s", label);
  
  context = (UnityWebappsLauncherContext *)user_data;

  real_add_action (context, label);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}

static gboolean
on_handle_set_urgent (UnityWebappsGenLauncher *launcher,
		      GDBusMethodInvocation *invocation,
		      gpointer user_data)
{
  UnityWebappsLauncherContext *context;
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Handling SetUrgent call");
  
  context = (UnityWebappsLauncherContext *)user_data;
  
  real_set_urgent (context);
  
  g_dbus_method_invocation_return_value (invocation, NULL);
  
  return TRUE;
}
  




static void
export_object (GDBusConnection *connection,
	       UnityWebappsLauncherContext *launcher_context)
{
  GError *error;
  
  launcher_context->launcher_service_interface = unity_webapps_gen_launcher_skeleton_new ();
  
  g_signal_connect (launcher_context->launcher_service_interface,
		    "handle-set-count",
		    G_CALLBACK (on_handle_set_count),
		    launcher_context);

  g_signal_connect (launcher_context->launcher_service_interface,
		    "handle-clear-count",
		    G_CALLBACK (on_handle_clear_count),
		    launcher_context);

  g_signal_connect (launcher_context->launcher_service_interface,
		    "handle-set-progress",
		    G_CALLBACK (on_handle_set_progress),
		    launcher_context);

  g_signal_connect (launcher_context->launcher_service_interface,
		    "handle-clear-progress",
		    G_CALLBACK (on_handle_clear_progress),
		    launcher_context);

  g_signal_connect (launcher_context->launcher_service_interface,
		    "handle-add-action",
		    G_CALLBACK (on_handle_add_action),
		    launcher_context);
  g_signal_connect (launcher_context->launcher_service_interface,
		    "handle-set-urgent",
		    G_CALLBACK (on_handle_set_urgent),
		    launcher_context);
  
  error = NULL;
  
  g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (launcher_context->launcher_service_interface),
				    connection,
				    UNITY_WEBAPPS_LAUNCHER_PATH,
				    &error);
  
  if (error != NULL)
    {
      g_error ("Error exporting Unity Webapps Launcher object: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Exported Launcher object");
}

UnityWebappsLauncherContext *
unity_webapps_launcher_context_new (GDBusConnection *connection, const gchar *desktop_id)
{
  UnityWebappsLauncherContext *launcher_context;

  UNITY_WEBAPPS_NOTE (LAUNCHER, "Creating new UnityWebappsLauncherContext object");
  
  launcher_context = g_malloc0 (sizeof (UnityWebappsLauncherContext));
  
  launcher_context->connection = g_object_ref (G_OBJECT (connection));
  
  export_object (connection, launcher_context);
  
  launcher_context->launcher_entry = unity_launcher_entry_get_for_desktop_id (desktop_id);
  
  launcher_context->menu_item = NULL;
  launcher_context->num_actions = 0;
  
  
  return launcher_context;
}

void
unity_webapps_launcher_context_free (UnityWebappsLauncherContext *launcher_context)
{
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Finalizing UnityWebappsLauncherContext object");
  g_object_unref (G_OBJECT (launcher_context->launcher_service_interface));
  g_object_unref (G_OBJECT (launcher_context->connection));
  
  g_object_unref (G_OBJECT (launcher_context->launcher_entry));
  
  if (launcher_context->menu_item != NULL)
    {
      g_object_unref (G_OBJECT (launcher_context->menu_item));
    }

  g_free (launcher_context);
}
