/*
 * unity-webapps-indicator-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>

#include "unity-webapps-indicator-context.h"
#include "unity-webapps-context-private.h"

#include "unity-webapps-sanitizer.h"

#include "unity-webapps-debug.h"

static void
_indicator_context_action_invoked (UnityWebappsGenIndicator *indicator,
				   const gchar *label,
				   gpointer user_data)
{
  UnityWebappsIndicatorContext *indicator_context;
  UnityWebappsIndicatorCallback callback;
  
  indicator_context = (UnityWebappsIndicatorContext *)user_data;
  
  callback = g_hash_table_lookup (indicator_context->menu_callbacks_by_name, label);
  
  if (callback != NULL)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Menu action invoked: %s", label);
      callback();
      return;
    }
  
  callback = g_hash_table_lookup (indicator_context->indicator_callbacks_by_name, label);
  
  if (callback != NULL)
    {
      UNITY_WEBAPPS_NOTE (INDICATOR, "Indicator action invoked: %s", label);
      callback();
      return;
    }

}


static void
presence_changed_notify (GObject *object,
			 GParamSpec *pspec,
			 gpointer user_data)
{
  UnityWebappsIndicatorContext *indicator_context = (UnityWebappsIndicatorContext *)user_data;
  GList *walk;
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Presence changed notification");
  
  for (walk = indicator_context->presence_callbacks_list; walk != NULL; walk = g_list_next (walk))
    {
      UnityWebappsPresenceCallback callback;
      
      callback = (UnityWebappsPresenceCallback)(walk->data);
      
      if (callback)
	callback();
    }
}

UnityWebappsIndicatorContext *
unity_webapps_indicator_context_new (GDBusConnection *connection, const gchar *name, GError **error)
{
  UnityWebappsIndicatorContext *context;
  
  g_return_val_if_fail (connection != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (g_dbus_is_name (name), NULL);
  
  context = g_malloc0 (sizeof (UnityWebappsIndicatorContext));
  context->indicator_rate = 0;
  context->presence_callbacks_list = NULL;
  
  context->menu_callbacks_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  context->indicator_callbacks_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  
  context->indicator_proxy = 
    unity_webapps_gen_indicator_proxy_new_sync (connection ,
						G_DBUS_PROXY_FLAGS_NONE,
						name,
						UNITY_WEBAPPS_INDICATOR_PATH, 
						NULL /* Cancellable */,
						error);
  
  if (error && (*error != NULL))
    {
      g_critical ("Error creating indicator context proxy object for %s: %s", name, (*error)->message);
      
      return NULL;
    }
  
  g_signal_connect (context->indicator_proxy, "action-invoked", G_CALLBACK (_indicator_context_action_invoked), context);
  g_signal_connect (context->indicator_proxy, "notify::presence", G_CALLBACK (presence_changed_notify), context);
  
  return context;
}

void 
unity_webapps_indicator_context_free (UnityWebappsIndicatorContext *context)
{
  g_return_if_fail (context != NULL);

  g_hash_table_destroy (context->menu_callbacks_by_name);
  g_hash_table_destroy (context->indicator_callbacks_by_name);
  g_list_free (context->presence_callbacks_list);
 

  g_object_unref (G_OBJECT (context->indicator_proxy));
  
  g_free (context);
}

static void
show_indicator_complete_callback (GObject *source_object,
				  GAsyncResult *res,
				  gpointer user_data)
{
  UnityWebappsGenIndicator *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_INDICATOR (source_object);
  
  error = NULL;
  
  unity_webapps_gen_indicator_call_show_indicator_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling ShowIndicator method of Indicator context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Received response, indicator successfully showed");
}

#define MAXIMUM_NAME_LENGTH 45

void
unity_webapps_indicator_show_indicator (UnityWebappsContext *context,
					const gchar *name)
{
  gchar *sanitized_name;

  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (name != NULL);
  
  if (unity_webapps_rate_check_indicator_rate_limit (context) == FALSE)
    {
      return;
    }
  
  sanitized_name = unity_webapps_sanitizer_limit_string_argument (name, MAXIMUM_NAME_LENGTH);
  unity_webapps_gen_indicator_call_show_indicator (context->priv->indicator_context->indicator_proxy,
						   sanitized_name,
						   NULL /* Cancellable */,
						   show_indicator_complete_callback,
						   context);

  g_free (sanitized_name);

  UNITY_WEBAPPS_NOTE (INDICATOR, "Showed indicator: %s", name);
}

static void
clear_indicator_complete_callback (GObject *source_object,
				  GAsyncResult *res,
				  gpointer user_data)
{
  UnityWebappsGenIndicator *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_INDICATOR (source_object);
  
  error = NULL;
  
  unity_webapps_gen_indicator_call_clear_indicator_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling ShowIndicator method of Indicator context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Received response, indicator successfully cleared");
}


void
unity_webapps_indicator_clear_indicator (UnityWebappsContext *context,
					 const gchar *name)
{
  gchar *sanitized_name;

  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (name != NULL);
  
  sanitized_name = unity_webapps_sanitizer_limit_string_argument (name, MAXIMUM_NAME_LENGTH);

  if (unity_webapps_rate_check_indicator_rate_limit (context) == FALSE)
    {
      return;
    }
  
  g_return_if_fail (name != NULL);
  
  unity_webapps_gen_indicator_call_clear_indicator (context->priv->indicator_context->indicator_proxy,
						    sanitized_name,
						    NULL /* Cancellable */,
						    clear_indicator_complete_callback,
						    context);
  
  g_free (sanitized_name);
  UNITY_WEBAPPS_NOTE (INDICATOR, "Cleared indicator: %s", name);
  
}

static void
add_action_complete_callback (GObject *source_object,
			      GAsyncResult *res,
			      gpointer user_data)
{
  UnityWebappsGenIndicator *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_INDICATOR (source_object);
  
  error = NULL;
  
  unity_webapps_gen_indicator_call_add_action_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling AddAction method of Indicator context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Received response, action succesfully added");
}

#define MAXIMUM_LABEL_LENGTH 60

void
unity_webapps_indicator_add_action (UnityWebappsContext *context,
				    const gchar *label,
				    UnityWebappsIndicatorCallback callback)
{
  gchar *sanitized_label;

  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (label != NULL);
  g_return_if_fail (callback != NULL);

  if (unity_webapps_rate_check_indicator_rate_limit (context) == FALSE)
    {
      return;
    }
  
  sanitized_label = unity_webapps_sanitizer_limit_string_argument (label, MAXIMUM_LABEL_LENGTH);

  unity_webapps_gen_indicator_call_add_action (context->priv->indicator_context->indicator_proxy,
					       sanitized_label,
					       NULL /* Cancellable */,
					       add_action_complete_callback,
					       context);
  
  
  // TODO: Analyze the error case with duplicates.
  g_hash_table_insert (context->priv->indicator_context->menu_callbacks_by_name, g_strdup (label), callback);
  
  g_free (sanitized_label);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Added action: %s", label);
}

static void
set_property_complete_callback (GObject *source_object,
			      GAsyncResult *res,
			      gpointer user_data)
{
  UnityWebappsGenIndicator *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_INDICATOR (source_object);
  
  error = NULL;
  
  unity_webapps_gen_indicator_call_set_property_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling SetProperty of Indicator context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Received response, property succesfully set");
}

#define MAXIMUM_PROPERTY_LENGTH 40
#define MAXIMUM_VALUE_LENGTH 1024


void 
unity_webapps_indicator_set_property (UnityWebappsContext *context, 
				      const gchar *name, 
				      const gchar *property, 
				      const gchar *value)
{
  gchar *sanitized_name, *sanitized_property, *sanitized_value;

  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (name != NULL);
  g_return_if_fail (property != NULL);
  g_return_if_fail (value != NULL);

  if (unity_webapps_rate_check_indicator_rate_limit (context) == FALSE)
    {
      return;
    }
  
  sanitized_name = unity_webapps_sanitizer_limit_string_argument (name, MAXIMUM_NAME_LENGTH);
  sanitized_property = unity_webapps_sanitizer_limit_string_argument (property, MAXIMUM_PROPERTY_LENGTH);
  sanitized_value = unity_webapps_sanitizer_limit_string_argument (value, MAXIMUM_VALUE_LENGTH);
  
  unity_webapps_gen_indicator_call_set_property (context->priv->indicator_context->indicator_proxy,
						 sanitized_name,
						 sanitized_property,
						 sanitized_value,
						 NULL /* Cancellable */,
						 set_property_complete_callback,
						 context);
  
  g_free (sanitized_name);
  g_free (sanitized_property);
  g_free (sanitized_value);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Setting property of indicator %s: %s->%s", name, property, value);
}


static void
set_property_icon_complete_callback (GObject *source_object,
				     GAsyncResult *res,
				     gpointer user_data)
{
  UnityWebappsGenIndicator *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_INDICATOR (source_object);
  
  error = NULL;
  
  unity_webapps_gen_indicator_call_set_property_icon_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling SetPropertyIcon of Indicator context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Received response, icon property succesfully set");
}



void 
unity_webapps_indicator_set_property_icon (UnityWebappsContext *context, 
				      const gchar *name, 
				      const gchar *property, 
				      const gchar *value)
{
  gchar *sanitized_name, *sanitized_property, *sanitized_value;

  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (name != NULL);
  g_return_if_fail (property != NULL);
  g_return_if_fail (value != NULL);

  if (unity_webapps_rate_check_indicator_rate_limit (context) == FALSE)
    {
      return;
    }

  sanitized_name = unity_webapps_sanitizer_limit_string_argument (name, MAXIMUM_NAME_LENGTH);
  sanitized_property = unity_webapps_sanitizer_limit_string_argument (property, MAXIMUM_PROPERTY_LENGTH);
  sanitized_value = unity_webapps_sanitizer_limit_string_argument (value, MAXIMUM_VALUE_LENGTH);
  
  
  unity_webapps_gen_indicator_call_set_property_icon (context->priv->indicator_context->indicator_proxy,
						      sanitized_name,
						      sanitized_property,
						      sanitized_value,
						      NULL /* Cancellable */,
						      set_property_icon_complete_callback,
						      context);

  g_free (sanitized_name);
  g_free (sanitized_property);
  g_free (sanitized_value);
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Setting icon property of indicator %s: %s->%s", name,
		      property, value);
}

void 
unity_webapps_indicator_set_callback (UnityWebappsContext *context, 
				      const gchar *name, 
				      UnityWebappsIndicatorCallback callback)
{
  UnityWebappsIndicatorContext *indicator_context;
  gchar *sanitized_name;

  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (name != NULL);
  g_return_if_fail (callback != NULL);
  
  // The hash table takes ownership of this.
  sanitized_name = unity_webapps_sanitizer_limit_string_argument (name, MAXIMUM_NAME_LENGTH);
		   
  
  indicator_context = context->priv->indicator_context;
  
  g_hash_table_insert (indicator_context->indicator_callbacks_by_name, 
		       (gpointer) sanitized_name, 
		       (gpointer) callback);
}

gchar *
unity_webapps_indicator_get_presence (UnityWebappsContext *context)
{
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context), NULL);
  
  if (unity_webapps_rate_check_indicator_rate_limit (context) == FALSE)
    {
      return FALSE;
    }
  return g_strdup (unity_webapps_gen_indicator_get_presence (context->priv->indicator_context->indicator_proxy));
}

void 
unity_webapps_indicator_on_presence_changed_callback (UnityWebappsContext *context,
						      UnityWebappsPresenceCallback callback)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (callback != NULL);
  
  context->priv->indicator_context->presence_callbacks_list = g_list_append (context->priv->indicator_context->presence_callbacks_list, callback);
}
