/*
 * unity-webapps-launcher-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>

#include "unity-webapps-launcher-context.h"
#include "unity-webapps-context-private.h"

#include "unity-webapps-sanitizer.h"

static void
_launcher_context_action_invoked (UnityWebappsGenLauncher *launcher,
				  const gchar *label,
				  gpointer user_data)
{
  UnityWebappsLauncherContext *launcher_context;
  UnityWebappsLauncherCallback callback;
  
  launcher_context = (UnityWebappsLauncherContext *)user_data;
  
  callback = g_hash_table_lookup (launcher_context->quicklist_callbacks_by_name, label);
  
  if (callback != NULL)
    {
      UNITY_WEBAPPS_NOTE (LAUNCHER, "Quicklist action invoked: %s", label);

      callback();
      return;
    }
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Quicklist action invoked, but we do not have a handler: %s", label);
}

UnityWebappsLauncherContext *
unity_webapps_launcher_context_new (GDBusConnection *connection, const gchar *name, GError **error)
{
  UnityWebappsLauncherContext *context;
  
  g_return_val_if_fail (connection != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (g_dbus_is_name (name), NULL);
  
  context = g_malloc0 (sizeof (UnityWebappsLauncherContext));
  context->launcher_rate = 0;
  
  context->launcher_proxy = 
    unity_webapps_gen_launcher_proxy_new_sync (connection ,
						   G_DBUS_PROXY_FLAGS_NONE,
						   name,
						   UNITY_WEBAPPS_LAUNCHER_PATH, 
						   NULL /* Cancellable */,
						   error);
  
  if (error && (*error != NULL))
    {
      g_critical ("Error creating launcher context proxy object for %s: %s", name, (*error)->message);
      
      return NULL;
    }
  
  context->quicklist_callbacks_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  
  g_signal_connect (context->launcher_proxy, "action-invoked", G_CALLBACK (_launcher_context_action_invoked), context);
  
  return context;
}

void 
unity_webapps_launcher_context_free (UnityWebappsLauncherContext *context)
{
  g_return_if_fail (context != NULL);
  g_object_unref (G_OBJECT (context->launcher_proxy));
  
  g_hash_table_destroy (context->quicklist_callbacks_by_name);
  
  g_free (context);
}

static void
set_count_complete_callback (GObject *source_object,
			     GAsyncResult *res,
			     gpointer user_data)
{
  UnityWebappsGenLauncher *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_LAUNCHER (source_object);
  
  error = NULL;
  
  unity_webapps_gen_launcher_call_set_count_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling SetCount method of Launcher context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Received response, count succesfully set");
}

void
unity_webapps_launcher_set_count (UnityWebappsContext *context,
				  gint count)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  
  if (unity_webapps_rate_check_launcher_rate_limit (context) == FALSE)
    {
      return;
    }
  
  unity_webapps_gen_launcher_call_set_count (context->priv->launcher_context->launcher_proxy,
					     count,
					     NULL /* Cancellable */,
					     set_count_complete_callback,
					     context);
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Setting count: %d", count);
}

static void
set_progress_complete_callback (GObject *source_object,
			     GAsyncResult *res,
			     gpointer user_data)
{
  UnityWebappsGenLauncher *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_LAUNCHER (source_object);
  
  error = NULL;
  
  unity_webapps_gen_launcher_call_set_progress_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling SetProgress of Launcher context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Received response, progress succesfully set");
}

void
unity_webapps_launcher_set_progress (UnityWebappsContext *context,
				     gdouble progress)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  
  if (unity_webapps_rate_check_launcher_rate_limit (context) == FALSE)
    {
      return;
    }
  
  unity_webapps_gen_launcher_call_set_progress (context->priv->launcher_context->launcher_proxy,
						progress,
						NULL /* Cancellable */,
						set_progress_complete_callback,
						context);

  UNITY_WEBAPPS_NOTE (LAUNCHER, "Setting progress: %f", progress);
}

static void
clear_count_complete_callback (GObject *source_object,
			     GAsyncResult *res,
			     gpointer user_data)
{
  UnityWebappsGenLauncher *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_LAUNCHER (source_object);
  
  error = NULL;
  
  unity_webapps_gen_launcher_call_clear_count_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling ClearCount method of Launcher context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Received response, count succesfully cleared");
}

void
unity_webapps_launcher_clear_count (UnityWebappsContext *context)
				  
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  
  if (unity_webapps_rate_check_launcher_rate_limit (context) == FALSE)
    {
      return;
    }
  
  unity_webapps_gen_launcher_call_clear_count (context->priv->launcher_context->launcher_proxy,
					       NULL /* Cancellable */,
					       clear_count_complete_callback,
					       context);
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Clearing count");
}

static void
clear_progress_complete_callback (GObject *source_object,
				  GAsyncResult *res,
				  gpointer user_data)
{
  UnityWebappsGenLauncher *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_LAUNCHER (source_object);
  
  error = NULL;
  
  unity_webapps_gen_launcher_call_clear_progress_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling ClearProgress method of Launcher context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Received response, progress succesfully cleared");
}

void
unity_webapps_launcher_clear_progress (UnityWebappsContext *context)
				  
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  
  if (unity_webapps_rate_check_launcher_rate_limit (context) == FALSE)
    {
      return;
    }
  
  unity_webapps_gen_launcher_call_clear_progress (context->priv->launcher_context->launcher_proxy,
						  NULL /* Cancellable */,
						  clear_progress_complete_callback,
						  context);
						
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Clearing progress");
}

static void
set_urgent_complete_callback (GObject *source_object,
			     GAsyncResult *res,
			     gpointer user_data)
{
  UnityWebappsGenLauncher *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_LAUNCHER (source_object);
  
  error = NULL;
  
  unity_webapps_gen_launcher_call_set_urgent_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling SetUrgent of Launcher context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Received response, urgent succesfully set");
}

void
unity_webapps_launcher_set_urgent (UnityWebappsContext *context)
				  
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  
  if (unity_webapps_rate_check_launcher_rate_limit (context) == FALSE)
    {
      return;
    }
  
  unity_webapps_gen_launcher_call_set_urgent (context->priv->launcher_context->launcher_proxy,
					      NULL /* Cancellable */,
					      set_urgent_complete_callback,
					      context);
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Setting urgent state");
}

static void
add_action_complete_callback (GObject *source_object,
			      GAsyncResult *res,
			      gpointer user_data)
{
  UnityWebappsGenLauncher *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_LAUNCHER (source_object);
  
  error = NULL;
  
  unity_webapps_gen_launcher_call_add_action_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling AddAction method of Launcher context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (INDICATOR, "Received response, action succesfully added");
}


#define MAXIMUM_LABEL_LENGTH 60

void
unity_webapps_launcher_add_action (UnityWebappsContext *context, const gchar *label, UnityWebappsLauncherCallback callback)
				  
{
  gchar *sanitized_label;

  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (callback != NULL);


  if (unity_webapps_rate_check_launcher_rate_limit (context) == FALSE)
    {
      return;
    }
  
  // The Hash Table takes ownership of this.
  sanitized_label = unity_webapps_sanitizer_limit_string_argument (label, MAXIMUM_LABEL_LENGTH);
  
  unity_webapps_gen_launcher_call_add_action (context->priv->launcher_context->launcher_proxy,
					      sanitized_label,
					      NULL /* Cancellable */,
					      add_action_complete_callback,
					      context);
  
  g_hash_table_insert (context->priv->launcher_context->quicklist_callbacks_by_name, sanitized_label, callback);
  
  UNITY_WEBAPPS_NOTE (LAUNCHER, "Adding action: %s", label);
}

