/*
 * unity-webapps-music-player-context.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>

#include "unity-webapps-music-player-context.h"

#include "unity-webapps-context-private.h"

#include "unity-webapps-sanitizer.h"

#include "unity-webapps-debug.h"

UnityWebappsMusicPlayerContext *
unity_webapps_music_player_context_new (GDBusConnection *connection, const gchar *name, GError **error)
{
  UnityWebappsMusicPlayerContext *context;
  
  g_return_val_if_fail (connection != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (g_dbus_is_name (name), NULL);
  
  context = g_malloc0 (sizeof (UnityWebappsMusicPlayerContext));
  context->music_player_rate = 0;
  
  context->music_player_proxy = 
    unity_webapps_gen_music_player_proxy_new_sync (connection ,
						   G_DBUS_PROXY_FLAGS_NONE,
						   name,
						   UNITY_WEBAPPS_MUSIC_PLAYER_PATH, 
						   NULL /* Cancellable */,
						   error);
  
  if (error && (*error != NULL))
    {
      g_critical ("Error creating music_player context proxy object for %s: %s", name, (*error)->message);
      
      return NULL;
    }
  
  return context;
}

void 
unity_webapps_music_player_context_free (UnityWebappsMusicPlayerContext *context)
{
  g_return_if_fail (context != NULL);

  g_object_unref (G_OBJECT (context->music_player_proxy));
  
  g_free (context);
}



// TODO: NOTE: REMOVE IN NEXT RELEASE
void
unity_webapps_music_player_init (UnityWebappsContext *context,
				 const gchar *title)
{
  return;
}

static void
set_track_complete_callback (GObject *source_object,
			     GAsyncResult *res,
			     gpointer user_data)
{
  UnityWebappsGenMusicPlayer *proxy;
  GError *error;
  
  proxy = UNITY_WEBAPPS_GEN_MUSIC_PLAYER (source_object);
  
  error = NULL;
  
  unity_webapps_gen_music_player_call_set_track_finish (proxy, res, &error);
  
  if (error != NULL)
    {
      g_warning ("Error calling SetTrack method of Indicator context: %s", error->message);
      
      g_error_free (error);
      
      return;
    }
  
  UNITY_WEBAPPS_NOTE (MUSIC_PLAYER, "Received response, track successfully set");
}


#define MAXIMUM_ARTIST_LENGTH 60
#define MAXIMUM_ALBUM_LENGTH 60
#define MAXIMUM_TITLE_LENGTH 60
#define MAXIMUM_ICON_URL_LENGTH 4096 /* For large data URIs...this seems reasonable */

void 
unity_webapps_music_player_set_track (UnityWebappsContext *context,
				      const gchar *artist,
				      const gchar *album,
				      const gchar *title,
				      const gchar *icon_url)
{
  gchar *sanitized_artist, *sanitized_album, *sanitized_title, *sanitized_icon_url;
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (title != NULL);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  
  sanitized_artist = unity_webapps_sanitizer_limit_string_argument (artist, MAXIMUM_ARTIST_LENGTH);
  sanitized_album = unity_webapps_sanitizer_limit_string_argument (album, MAXIMUM_ALBUM_LENGTH);
  sanitized_title = unity_webapps_sanitizer_limit_string_argument (title, MAXIMUM_TITLE_LENGTH);
  sanitized_icon_url = unity_webapps_sanitizer_limit_string_argument (icon_url, MAXIMUM_ICON_URL_LENGTH);
  
  unity_webapps_gen_music_player_call_set_track (context->priv->music_player_context->music_player_proxy,
						 sanitized_artist, 
						 sanitized_album,
						 sanitized_title,
						 sanitized_icon_url,
						 NULL /* Cancellable */,
						 set_track_complete_callback,
						 context);
  
  g_free (sanitized_artist);
  g_free (sanitized_album);
  g_free (sanitized_title);
  g_free (sanitized_icon_url);
  
  UNITY_WEBAPPS_NOTE (MUSIC_PLAYER, "Setting track: %s by %s from %s", title, artist, album);
}

static void
_do_callback (UnityWebappsGenMusicPlayer *player, gpointer user_data)
{
  UnityWebappsMusicPlayerCallback callback = (UnityWebappsMusicPlayerCallback )user_data;
  
  if (callback)
    callback();
}

void 
unity_webapps_music_player_on_play_pause_callback (UnityWebappsContext *context, 
						   UnityWebappsMusicPlayerCallback callback)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (callback != NULL);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  
  g_return_if_fail (callback != NULL);
  g_signal_connect (context->priv->music_player_context->music_player_proxy, "play_pause", G_CALLBACK (_do_callback), callback);
}
void 
unity_webapps_music_player_on_previous_callback (UnityWebappsContext *context, 
						 UnityWebappsMusicPlayerCallback callback)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (callback != NULL);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  g_return_if_fail (callback != NULL);
  g_signal_connect (context->priv->music_player_context->music_player_proxy, "previous", G_CALLBACK (_do_callback), callback);
}
void 
unity_webapps_music_player_on_next_callback (UnityWebappsContext *context, 
					     UnityWebappsMusicPlayerCallback callback)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));
  g_return_if_fail (callback != NULL);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  g_return_if_fail (callback != NULL);
  g_signal_connect (context->priv->music_player_context->music_player_proxy, "next", G_CALLBACK (_do_callback), callback);
}

UnityWebappsMusicPlayerPlaybackState
unity_webapps_music_player_get_playback_state (UnityWebappsContext *context)
{
  g_return_val_if_fail (context != NULL, 0);
  g_return_val_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context), 0);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return FALSE;
    }
  
  return (UnityWebappsMusicPlayerPlaybackState) unity_webapps_gen_music_player_get_playback_state (context->priv->music_player_context->music_player_proxy);
}

gboolean
unity_webapps_music_player_get_can_go_previous (UnityWebappsContext *context)
{
  g_return_val_if_fail (context != NULL, FALSE);
  g_return_val_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context), FALSE);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return FALSE;
    }
  return unity_webapps_gen_music_player_get_can_go_previous (context->priv->music_player_context->music_player_proxy);
}

gboolean
unity_webapps_music_player_get_can_go_next (UnityWebappsContext *context)
{  
  g_return_val_if_fail (context != NULL, FALSE);
  g_return_val_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context), FALSE);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return FALSE;
    }
  return unity_webapps_gen_music_player_get_can_go_next (context->priv->music_player_context->music_player_proxy);
}

gboolean
unity_webapps_music_player_get_can_play (UnityWebappsContext *context)
{
  g_return_val_if_fail (context != NULL, FALSE);
  g_return_val_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context), FALSE);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return FALSE;
    }
  return unity_webapps_gen_music_player_get_can_play (context->priv->music_player_context->music_player_proxy);
}

gboolean
unity_webapps_music_player_get_can_pause (UnityWebappsContext *context)
{
  g_return_val_if_fail (context != NULL, FALSE);
  g_return_val_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context), FALSE);

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return FALSE;
    }
  return unity_webapps_gen_music_player_get_can_pause (context->priv->music_player_context->music_player_proxy);
}

void 
unity_webapps_music_player_set_playback_state (UnityWebappsContext *context, UnityWebappsMusicPlayerPlaybackState state)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  g_dbus_proxy_set_cached_property (G_DBUS_PROXY (context->priv->music_player_context->music_player_proxy), 
				    "PlaybackState",
				    g_variant_new("i",
						  state,
						  NULL));
  unity_webapps_gen_music_player_set_playback_state (context->priv->music_player_context->music_player_proxy, (guint) state);
}

void
unity_webapps_music_player_set_can_go_previous (UnityWebappsContext *context, gboolean can_go_previous)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  g_dbus_proxy_set_cached_property (G_DBUS_PROXY (context->priv->music_player_context->music_player_proxy), 
				    "CanGoPrevious",
				    g_variant_new("b",
						  can_go_previous,
						  NULL));
  unity_webapps_gen_music_player_set_can_go_previous (context->priv->music_player_context->music_player_proxy, can_go_previous);
}

void
unity_webapps_music_player_set_can_go_next (UnityWebappsContext *context, gboolean can_go_next)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  g_dbus_proxy_set_cached_property (G_DBUS_PROXY (context->priv->music_player_context->music_player_proxy), 
				    "CanGoNext",
				    g_variant_new("b",
						  can_go_next,
						  NULL));
  unity_webapps_gen_music_player_set_can_go_next (context->priv->music_player_context->music_player_proxy, can_go_next);
}

void
unity_webapps_music_player_set_can_play (UnityWebappsContext *context, gboolean can_play)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  g_dbus_proxy_set_cached_property (G_DBUS_PROXY (context->priv->music_player_context->music_player_proxy), 
				    "CanPlay",
				    g_variant_new("b",
						  can_play,
						  NULL));
  unity_webapps_gen_music_player_set_can_play (context->priv->music_player_context->music_player_proxy, can_play);
}

void
unity_webapps_music_player_set_can_pause (UnityWebappsContext *context, gboolean can_pause)
{
  g_return_if_fail (context != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_CONTEXT (context));

  if (unity_webapps_rate_check_music_player_rate_limit (context) == FALSE)
    {
      return;
    }
  g_dbus_proxy_set_cached_property (G_DBUS_PROXY (context->priv->music_player_context->music_player_proxy), 
				    "CanPause",
				    g_variant_new("b",
						  can_pause,
						  NULL));
  unity_webapps_gen_music_player_set_can_pause (context->priv->music_player_context->music_player_proxy, can_pause);
}
