/*
 * unity-webapps-permissions.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include "unity-webapps-permissions.h"
#include <gio/gio.h>

/*
 * Add a domain to the list of "dontask" domains which should be blacklisted from requesting permissions
 */
void
unity_webapps_permissions_dontask_domain (const gchar *domain)
{
  GSettings *settings;
  gchar **dontask_domains, **new_dontask_domains;
  int i, length;
  
  settings = g_settings_new("com.canonical.unity.webapps");
  dontask_domains = g_settings_get_strv(settings, "dontask-domains");
  
  length = g_strv_length (dontask_domains);
  
  new_dontask_domains = g_malloc (sizeof(gchar *) * (length + 2));
  
    for (i = 0; i < length; i++)
    {
      new_dontask_domains[i] = g_strdup (dontask_domains[i]);
    }
  
  new_dontask_domains[length] = g_strdup (domain);
  new_dontask_domains[length+1] = NULL;
  
  g_settings_set_strv(settings, "dontask-domains", (const gchar * const*)new_dontask_domains);
  
    g_object_unref (G_OBJECT (settings));

    g_strfreev (dontask_domains);
    g_strfreev (new_dontask_domains);
}

/*
 * Add a domain to the list of domains which have been granted permissions
 */
void
unity_webapps_permissions_allow_domain (const gchar *domain)
{
  GSettings *settings;
  gchar **allowed_domains, **new_allowed_domains;
  int i, length;
  
  settings = g_settings_new("com.canonical.unity.webapps");
  allowed_domains = g_settings_get_strv(settings, "allowed-domains");
  
  length = g_strv_length (allowed_domains);
  
  new_allowed_domains = g_malloc (sizeof(gchar *) * (length + 2));
  
    for (i = 0; i < length; i++)
    {
      new_allowed_domains[i] = g_strdup (allowed_domains[i]);
    }
  
  new_allowed_domains[length] = g_strdup (domain);
  new_allowed_domains[length+1] = NULL;
  
  g_settings_set_strv(settings, "allowed-domains", (const gchar * const*)new_allowed_domains);
  
    g_object_unref (G_OBJECT (settings));

    g_strfreev (allowed_domains);
    g_strfreev (new_allowed_domains);
}

/*
 * Checks if a domain is in the list of domains allowed permissions.
 */
gboolean
unity_webapps_permissions_get_domain_allowed (const gchar *domain)
{
  GSettings *settings;
  gchar **allowed_domains;
  int i;
  gboolean ret;
  
  settings = g_settings_new ("com.canonical.unity.webapps");

  allowed_domains = g_settings_get_strv(settings, "allowed-domains");
  
  if (!allowed_domains)
    {
      g_object_unref (G_OBJECT (settings));
      return FALSE;
    }
  
  ret = FALSE;

  for (i = 0; allowed_domains[i] != NULL; i++)
    {
      if (g_strcmp0(allowed_domains[i], domain) == 0)
	{
	  ret = TRUE;
	  break;
	}
    }
  
  g_object_unref (G_OBJECT (settings));
  g_strfreev (allowed_domains);
  
  return ret;
}

/*
 * Checks if a domain is in the list of blacklisted dontask domains
 */
gboolean
unity_webapps_permissions_get_domain_dontask (const gchar *domain)
{
  GSettings *settings;
  gchar **dontask_domains;
  int i;
  gboolean ret;
  
  settings = g_settings_new ("com.canonical.unity.webapps");

  dontask_domains = g_settings_get_strv(settings, "dontask-domains");
  
  if (!dontask_domains)
    {
      g_object_unref (G_OBJECT (settings));
      return FALSE;
    }
  
  ret = FALSE;

  for (i = 0; dontask_domains[i] != NULL; i++)
    {
      if (g_strcmp0(dontask_domains[i], domain) == 0)
	{
	  ret = TRUE;
	  break;
	}
    }
  
  g_object_unref (G_OBJECT (settings));
  g_strfreev (dontask_domains);
  
  return ret;
}
