/*
 * unity-webapps-service.c
 * Copyright (C) Canonical LTD 2011
 *
 * Author: Robert Carr <racarr@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

#include <stdio.h>

#include "unity-webapps-service.h"

struct _UnityWebappsServicePrivate {
  GDBusConnection *connection;
  UnityWebappsGenService *service_proxy;
};

G_DEFINE_TYPE(UnityWebappsService, unity_webapps_service, G_TYPE_OBJECT)

#define UNITY_WEBAPPS_SERVICE_GET_PRIVATE(object) (G_TYPE_INSTANCE_GET_PRIVATE ((object), UNITY_WEBAPPS_TYPE_SERVICE, UnityWebappsServicePrivate))

static void
unity_webapps_service_finalize (GObject *object)
{
  UnityWebappsService *service;
  
  service = UNITY_WEBAPPS_SERVICE (object);

  g_object_unref (G_OBJECT (service->priv->service_proxy));
  g_object_unref (G_OBJECT (service->priv->connection));
}

static void
unity_webapps_service_class_init (UnityWebappsServiceClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  
  object_class->finalize = unity_webapps_service_finalize;
  
  g_type_class_add_private (object_class, sizeof(UnityWebappsServicePrivate));
}

static void
unity_webapps_service_init (UnityWebappsService *service)
{
  GError *error;

  service->priv = UNITY_WEBAPPS_SERVICE_GET_PRIVATE (service);

#ifdef UNITY_WEBAPPS_ENABLE_DEBUG
  unity_webapps_debug_initialize_flags ();
#endif
  
  error = NULL;

  service->priv->connection = g_bus_get_sync (G_BUS_TYPE_SESSION,
					 NULL /*Cancellable*/,
					 &error /*Error*/);
  
  if (error != NULL)
    {
      g_critical ("Failed to open session bus! %s", error->message);

      g_error_free (error);
      return;
    }
  
  service->priv->service_proxy = unity_webapps_gen_service_proxy_new_sync (service->priv->connection,
									     G_DBUS_PROXY_FLAGS_NONE,
									     UNITY_WEBAPPS_SERVICE_NAME,
									     UNITY_WEBAPPS_SERVICE_PATH,
									     NULL,
									     &error);
  
  if (error != NULL)
    {
      g_critical ("Failed to create service proxy. %s", error->message);
      g_error_free (error);
      
      return;
    }
  
}

UnityWebappsService *
unity_webapps_service_new ()
{
  return g_object_new (UNITY_WEBAPPS_TYPE_SERVICE, NULL);
}

GDBusConnection *
unity_webapps_service_get_connection (UnityWebappsService *service)
{
  return service->priv->connection;
}

UnityWebappsGenService *
unity_webapps_service_get_proxy (UnityWebappsService *service)
{
  return service->priv->service_proxy;
}

void
unity_webapps_service_activate_application (UnityWebappsService *service, 
					    const gchar *name, 
					    const gchar *domain)
{
  GError *error;

  g_return_if_fail (service != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_SERVICE (service));
  g_return_if_fail (name != NULL);
  g_return_if_fail (domain != NULL);
  
  
  error = NULL;

  unity_webapps_gen_service_call_activate_application_sync (service->priv->service_proxy, name, domain, NULL, &error);
  
  if (error != NULL)
    {
      g_warning ("Failure calling ActivateApplication (%s,%s): %s", name, domain, error->message);
      g_error_free (error);
      
      return;
    }
  
}

typedef struct _unity_webapps_service_notify_data {
  gpointer user_data;
  UnityWebappsServiceContextNotifyCallback callback;
  UnityWebappsService *service;
} unity_webapps_service_notify_data;

static void
on_notify_context_appeared (UnityWebappsGenService *gen_service,
			    const gchar *name,
			    gpointer user_data)
{
  unity_webapps_service_notify_data *data;
  
  data = (unity_webapps_service_notify_data *)user_data;
  
  data->callback(data->service, name, data->user_data);
}

void
unity_webapps_service_on_context_appeared (UnityWebappsService *service,
					   UnityWebappsServiceContextNotifyCallback callback,
					   gpointer user_data)
{
  unity_webapps_service_notify_data *data;
  
  g_return_if_fail (service != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_SERVICE (service));
  g_return_if_fail (callback != NULL);
  
  data = g_malloc0 (sizeof (unity_webapps_service_notify_data));
  data->user_data = user_data;
  data->callback = callback;
  data->service = service;
  
  g_signal_connect (service->priv->service_proxy,
		    "context-appeared",
		    G_CALLBACK (on_notify_context_appeared),
		    data);
}



static void
on_notify_context_vanished (UnityWebappsGenService *gen_service,
			    const gchar *name,
			    gpointer user_data)
{
  unity_webapps_service_notify_data *data;
  
  data = (unity_webapps_service_notify_data *)user_data;
  
  data->callback(data->service, name, data->user_data);
}

void
unity_webapps_service_on_context_vanished (UnityWebappsService *service,
					   UnityWebappsServiceContextNotifyCallback callback,
					   gpointer user_data)
{
  unity_webapps_service_notify_data *data;
  
  g_return_if_fail (service != NULL);
  g_return_if_fail (UNITY_WEBAPPS_IS_SERVICE (service));
  g_return_if_fail (callback != NULL);
  
  data = g_malloc0 (sizeof (unity_webapps_service_notify_data));
  data->user_data = user_data;
  data->callback = callback;
  data->service = service;
  
  g_signal_connect (service->priv->service_proxy,
		    "context-vanished",
		    G_CALLBACK (on_notify_context_vanished),
		    data);
}

gchar **
unity_webapps_service_list_contexts (UnityWebappsService *service)
{
  gchar **out_contexts;
  GError *error;
  
  g_return_val_if_fail (service != NULL, NULL);
  g_return_val_if_fail (UNITY_WEBAPPS_IS_SERVICE (service), NULL);
  
  error = NULL;
  
  unity_webapps_gen_service_call_list_contexts_sync (service->priv->service_proxy,
						     &out_contexts,
						     NULL /* Cancellable */,
						     &error);
  
  if (error != NULL)
    {
      g_warning ("Failure calling ListContexts: %s", error->message);
      g_error_free (error);
      
      return NULL;
    }
  
  return out_contexts;
}
